# .
# project.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
require 'zlib'
require 'yaml'
require 'lib/httpparser'
require 'lib/qShapers'
require 'lib/core/session'
require 'lib/core/passive_check'
require 'lib/core/scanner'

module Watobo
  module Core
    class Conversation

      attr_accessor :file
      def id()
        # must be defined
      end

      def copyRequest()
        # req_copy = []
        # self.request.each do |line|
        #   req_copy.push line.clone
        # end
        copy = Utils.copyObject(@request)
        # now extend the new request with the Watobo mixins
        copy.extend Watobo::Parser::Url
        copy.extend Watobo::Parser::Web10
        copy.extend Watobo::Framework::Shaper::Web10
        return copy
      end

      private

      def extendRequest
        @request.extend Watobo::Framework::Shaper::Web10
        @request.extend Watobo::Parser::Web10
        @request.extend Watobo::Parser::Url
      end

      def extendResponse
        @response.extend Watobo::Parser::Web10
      end

      def initialize(request, response)
        @request = request
        @response = response
        @file = nil

        extendRequest()
        extendResponse()

      end

    end

    class Chat < Conversation
      attr :request
      attr :response
      attr :settings

      @@numChats = 0
      @@max_id = 0

      @@lock = Mutex.new

      public
      def resetCounters()
        @@numChats = 0
        @@max_id = 0
      end

      def tested?()
        return false unless @settings.has_key?(:tested)
        return @settings[:tested]
      end

      def tested=(truefalse)
        @settings[:tested] = truefalse
      end

      def tstart()
        @settings[:tstart]
      end

      def tstop()
        @settings[:tstop]
      end

      def id()
        @settings[:id]
      end

      def comment=(c)
        @settings[:comment] = c
      end

      def comment()
        @settings[:comment]
      end

      def source()
        @settings[:source]
      end

      # INITIALIZE ( request, response, prefs )
      # prefs:
      #   :source - source of request/response CHAT_SOURCE
      #   :id     - an initial id, if no id is given it will be set to the @@max_id, if id == 0 counters will be ignored.
      #   :start  - starting time of request format is Time.now.to_f
      #   :stop   - time of loading response has finished
      #   :
      def initialize(request, response, prefs = {})

        begin
          super(request, response)

          @settings = {
            :source => CHAT_SOURCE_UNDEF,
            :id => -1,
            :start => 0,
            :stop => -1,
            :comment => '',
            :tested => false
          }

          @settings.update prefs
          #  puts @settings[:id].to_s

          @@lock.synchronize{
            # enter critical section here ???
            if @settings[:id] > @@max_id
              @@max_id = @settings[:id]
            elsif @settings[:id] < 0
              @@max_id += 1
              @settings[:id] = @@max_id
            end
            @@numChats += 1
            # @comment = ''
            # leafe critical section here ???
          }

        rescue => bang
          puts bang
          puts bang.backtrace
        end
      end

    end

    class Finding < Conversation

      @@numFindings = 0
      @@max_id = 0

      @@lock = Mutex.new

      attr :details
      attr :request
      attr :response
      def resetCounters()
        @@numFindings = 0
        @@max_id = 0
      end

      def id()
        @details[:id]
      end

      def initialize(request, response, details=nil)
        super(request, response)
        @details = {
          :id => -1,
          :comment => '',
          :false_positive => false    # FalsePositive
        }

        @details.update details if !details.nil? and details.is_a? Hash

        @@lock.synchronize{
          # enter critical section here ???
          if @details[:id] > 0 and @details[:id] > @@max_id
            @@max_id = @details[:id]
          elsif @details[:id] < 0
            @@max_id += 1
            @details[:id] = @@max_id
          end
          @@numFindings += 1

        }
        extendRequest()
        extendResponse()

      end

    end

    class Project
      # SessionManager
      #
      #     class SessionManager < Watobo::Core::Session
      #       def initialize(project)
      #         super(project.session )

      #       end
      #       def runLogin
      #         print "[L"
      #        @project.loginscript_ids.each do |chatid|
      #            test_req = @project.getChat(chatid).copyRequest
      #          request, response = doRequest(test_req, :update_sids => true, :update_session => true, :update_contentlength => true)
      #          end

      #      end
      #    end
      attr :chats
      attr_accessor :findings

      attr :date

      attr_accessor :settings

      # attr_accessor :numRunningChecks

      # attr_accessor :valid_sids
      ##   attr_accessor :loginscript_ids
      attr_accessor :active_checks
      attr_accessor :passive_checks
      attr_accessor :plugins
      attr_accessor :excluded_chats

      attr_accessor :intercept_request
      attr_accessor :intercept_response
      attr_accessor :interceptor

      attr :target_filter

      include Watobo::Utils
      include Watobo::Core
      def subscribe(event, &callback)
        (@event_dispatcher_listeners[event] ||= []) << callback
      end

      def notify(event, *args)
        if @event_dispatcher_listeners[event]
          @event_dispatcher_listeners[event].each do |m|
            m.call(*args) if m.respond_to? :call
          end
        end
      end

      def getLoginChats()
        login_chats = []
        @settings[:login_chat_ids].each do |cid|
          chat = getChat(cid)
          login_chats.push chat if chat
        end
        login_chats
      end

      def getWwwAuthentication()
        @settings[:www_auth]
      end

      def getLoginChatIds()
        #p @settings[:login_chat_ids]
        # p @settings.to_yaml
        @settings[:login_chat_ids]
      end

     

      def setLoginChatIds(ids)
        @settings[:login_chat_ids] = ids if ids.is_a? Array
      end

      def getSidPatterns
        @settings[:sid_patterns]
      end

      def setProxyOptions(proxy_prefs)
        @settings[:forwarding_proxy] = proxy_prefs

        @sessionController.addProxy(getCurrentProxy())
      end

      # gives the currently selected proxy
      # format <host>:<port>
      def getCurrentProxy()
        c_proxy = nil
        begin
          name = @settings[:forwarding_proxy][:name]
          cproxy = @settings[:forwarding_proxy][:list][name]
          return cproxy
        rescue
          puts "! no proxy settings available"
        end
        return nil
      end

      def setSidPatterns(sp)
        @settings[:sid_patterns] = sp if sp.is_a? Array
      end

      def getLogoutSignatures
        @settings[:logout_signatures]
      end

      def getCSRFPatterns
        @settings[:csrf_patterns]
      end


      # setCSRFRequest
      # =Parameters
      # request: test request which requires csrf handling
      # ids:      csrf request ids of current conversation
      # patterns: csrf patterns for identifiying and updating tokens
      def setCSRFRequest(request, ids, patterns=[])
        urh = uniqueRequestHash(request)
        puts ids.class
        @settings[:csrf_request_ids][urh] = ids
        @settings[:csrf_patterns].concat patterns unless patterns.empty?
        @settings[:csrf_patterns].uniq!
        notify(:settings_changed)
      end
      
      def getCSRFRequestIDs(request)
        urh = uniqueRequestHash(request)
        ids = @settings[:csrf_request_ids][urh]
         # puts "* found csrf req ids #{ids}"
        ids = [] if ids.nil?
          ids                
      end

      def setLogoutSignatures(ls)
        @settings[:logout_signatures] = ls if ls.is_a? Array
      end

      # Helper function to get all necessary preferences for starting a scan.
      def getScanPreferences()
        settings = {
          :smart_scan => @settings[:smart_scan],
          :non_unique_parms => @settings[:non_unique_parms],
          :excluded_parms => @settings[:excluded_parms],
          :sid_patterns => @settings[:sid_patterns],
          :csrf_patterns => @settings[:csrf_patterns],
          :run_passive_checks => false,
          :login_chat_ids => [],
          :proxy => getCurrentProxy(),
          :login_chats => getLoginChats(),
          # :scanlog_dir => scanLogDirectory(),
          :max_parallel_checks => @settings[:max_parallel_checks],
          :logout_signatures => @settings[:logout_signatures],
          :custom_error_patterns => @settings[:custom_error_patterns],
          :scan_session => self.object_id,
          :www_auth => @settings[:www_auth]
        }
        return settings
      end

      # returns a project/session specific ID needed for synchronising Sessions
      def getSessionID()
        sid = @settings[:project_name] + @settings[:session_name]
        return sid
      end

      def scanLogDirectory()
        sld = @settings[:scanlog_dir]
        return sld if File.exist?(sld)
        return ''
      end

      def scanLogDirectory=(dir)
        @settings[:scanlog_dir] = dir
      end

      def getScanPolicy()
        @settings[:policy]
      end

      def uniqueRequestHash(request)
        begin
          parms = request.site + request.method + request.path
          request.get_parm_names.sort.each do |p|
            if @settings[:non_unique_parms].include?(p) then
              parms += p + request.get_parm_value(p)
            else
              parms += p
            end

          end
          request.post_parm_names.sort.each do |p|
            if @settings[:non_unique_parms].include?(p) then
              parms += p + request.post_parm_value(p)
            else
              parms += p
            end

          end
          return Digest::MD5.hexdigest(parms)
        rescue
          return nil
        end
      end

      def last_chatid
        if @chats.length > 0
          return @chats.last.id
        else
          return 0
        end
      end

      def extendRequest(request)
        request.extend Watobo::Framework::Shaper::Web10
        request.extend Watobo::Parser::Web10
        request.extend Watobo::Parser::Url
      end

      def extendResponse(response)
        response.extend Watobo::Framework::Shaper::Web10
        response.extend Watobo::Parser::Web10
        response.extend Watobo::Parser::Url

      end

      def updateSettings(new_settings)
        new_settings.keys.each do |k|
          @settings[k] = new_settings[k]
        end
      end

      def listActiveCheckClasses()
        check_classes = []
        ignore = ['.','..','.svn']
        # ignore_modules = ["http_methods.rb"]
        ignore_modules = []

        active_path = File.join($watobo_base, @settings[:module_dir],"active")
        begin
          Dir.chdir(active_path)
          puts "Active Modules Directory: \"#{Dir.getwd}\""
        rescue StandardError => bang
          puts bang
          puts "\nError changing into source directory \"#{active_path}\"!"
          exit
        end

        Dir.foreach(".") do |group|

          if File.ftype(group) == "directory" and !ignore.include?(group)

            Dir.foreach("./#{group}") do |modules|
              if File.ftype("./#{group}/#{modules}") == "file" and modules =~ /.rb$/
                begin
                  module_file = File.join(active_path, group, modules)
                  require module_file
                  group_class = group.slice(0..0).upcase + group.slice(1..-1).downcase
                  module_class =modules.slice(0..0).upcase + modules.slice(1..-1).downcase
                  module_class.sub!(".rb","")
                  class_constant = Watobo.class_eval("Watobo::Modules::Active::#{group_class}::#{module_class}")

                  check_classes.push class_constant

                rescue => bang
                  puts bang
                  notify(:logger, LOG_DEBUG, "problems loading module: #{@settings[:module_path]}/active/#{group}/#{modules}")
                end

              end
            end
          end

        end
        return check_classes
      end

      def findChats(site, opts={})
        o = {
          :dir => "",
          #:file => nil,
          :method => nil,
          :max_count => 0
        }
        o.update opts
        o[:dir].strip!
        o[:dir].gsub!(/^\//,"")

        matches = []
        @chats.each do |c|
          if c.request.site == site then
            matches.push c if o[:dir] == c.request.dir
          end
          return matches if o[:max_count] > 0 and matches.length >= o[:max_count]
        end
        return matches
      end

      def getChat(chatid)
        #TODO: Improvement for faster results => Fuzzy-Logic
        @chats.each do |c|
          if c.id.to_s == chatid.to_s then
            return c
          end
        end
        return nil
      end

      def projectName
        @settings[:project_name]
      end

      def sessionName
        @settings[:session_name]
      end

      def interceptPort
        @settings[:project_name]
      end

      # runs passive checks on specific chat
      def runPassiveChecks(chat)
        tlist = []
        @passive_checks.each do |test_module|
          tlist << Thread.new(chat, test_module) {|c,m|
            m.do_test(c)
          }
        end
        return tlist
      end

      def addChat(chat, prefs={})
        begin
          if chat.request.host then
            chats.push chat

            options = { :run_passive_checks => true }
            options.update prefs

            #  p "*RP*"
            # puts "Running passive checks on chat #{chat.id}" if options[:run_passive_checks] == true
            runPassiveChecks(chat) if options[:run_passive_checks] == true

            #@interface.addChat(self, chat) if @interface
            notify(:new_chat, chat)

            if chat.id != 0 then
              if @conversations_path != "" then

                chat_file = File.join("#{@conversations_path}", "#{chat.id}-chat")
                chat_data = {
                  :request => chat.request.map{|x| x.inspect},
                  :response => chat.response.map{|x| x.inspect},
                }

                chat_data.update(chat.settings)
                if not File.exists?(chat_file) then
                  File.open(chat_file, "w") { |fh|
                    YAML.dump(chat_data, fh)
                  }
                  chat.file = chat_file
                end

              end
            else
              puts "!!! Could not add chat #{chat.id}"
            end
          end
          # p "!P!"
        rescue => bang
          puts bang
          puts bang.backtrace
        end
      end

      def runLogin
        @sessionMgr.runLogin(loginChats)
      end

      # TODO: check
      def createScanner_UNUSED()
        scan_modules = []
        scan_chats = []
        uniqueRequests = Hash.new
        notify(:logger, LOG_INFO, "*** SETTING UP SCANNER ***")
        notify(:logger, LOG_DEBUG, "Targets:")

        @active_checks.each do |test_module|
          if not test_module.enabled? then
            puts "+ skip disabled module #{test_module.info[:check_name]}"
            next
          end
          scan_modules.push test_module
        end

        @chats.each do |chat|
          if @excluded_chats.include?(chat.id) then
            puts "* [#{chat.id}] skip excluded chat"
            next
          end
          if @excluded_sites.include?(chat.request.site) then
            puts "* [#{chat.id}] skip excluded site: #{chat.request.site}"
            next
          end
          if @settings[:target_sites].include?(chat.request.site) then
            uh = uniqueRequestHash(chat.request)
            if not uniqueRequests.has_key?(uh)
              scan_chats.push chat
              puts chat.request.first
              uniqueRequests[uh] = chat
            else
              puts "* skipping chat[#{chat.id}] - hash(#{uh}) already exists"
            end
          end
        end

        notify(:logger, LOG_DEBUG, "Selected Modules (#{scan_modules.length}):")
        scan_modules.each do |mod|
          notify(:logger, LOG_DEBUG, "+ #{mod.info[:check_name]}")
        end
        notify(:logger, LOG_DEBUG, "Number of Chats: #{scan_chats.length}")
        scanner = Scanner.new(self, @interface, scan_modules, scan_chats)
      end

      def run_passive_checks
        #
        # PASSIVE CHECKS
        #
        notify(:logger, LOG_DEBUG, "====== starting passive checks ======")
        @passive_checks.each do |test_module|

          begin

            notify(:logger, LOG_DEBUG, "running checks with module (#{test_module.class})")
            @chats.each do |chat|

              @dashboard.update_status(test_module, @chats.index(chat))if @dashboard

              test_module.do_test(chat)

            end

          rescue => bang
            notify(:logger, LOG_INFO, "problem in module (#{test_module.class})\n#{bang}")
          end
          notify(:logger, LOG_DEBUG, "====== finished passive checks ======")
        end
      end

      # def getProjectSettings()
      # project_settings = Hash.new
      # project_settings.update @active_project[:forwarding_proxy]
      #Watobo::Utils.save_settings(@project_file, 0, "project_settings", project_settings)
      #Watobo::Utils.save_settings(@project_file, @settings)
      # end
      def has_scope?()
        return false if @settings[:scope].nil?
        @settings[:scope].each_key do |k|
          return true if @settings[:scope][k][:enabled] == true
        end
        return false
      end

      def scope
        @settings[:scope]
      end

      def scope=(scope)
        @settings[:scope] = scope
      end

      def setScope(scope)
        @settings[:scope] = scope
      end

      def setWwwAuthentication(www_auth)
        @settings[:www_auth] = www_auth
      #  puts "* setWwwAuth:"
      #  puts @settings[:www_auth].to_yaml
      end

      def setCSRFPatterns(patterns)
        @settings[:csrf_patterns] = patterns
      end

      def addToScope(site)
        return false if !@settings[:scope][site].nil?

        scope_details = {
          :site => site,
          :enabled => true,
          :root_path => '',
          :excluded_paths => [],
        }

        @settings[:scope][site] = scope_details
        return true
      end

      def addFinding(finding, opts={})
        # p "* add finding #{finding.details[:fid]}"
        unless @findings.has_key?(finding.details[:fid])
          begin
            @findings[finding.details[:fid]] = finding
            #@interface.addFinding(new_finding)
            #   puts "* new finding"
            notify(:new_finding, finding)

            finding_file = File.join("#{@findings_path}", "#{finding.id}-finding")
            if not File.exists?(finding_file) then

              finding_data = {
                :request => finding.request.map{|x| x.inspect},
                :response => finding.response.map{|x| x.inspect},
                :details => Hash.new
              }
              finding_data[:details].update(finding.details)

              if not File.exists?(finding_file) then
                fh = File.new(finding_file, "w+b")
                fh.print YAML.dump(finding_data)
                fh.close
              end
            end

          rescue => bang
            puts "!!!ERROR: #{Module.nesting[0].name}"
            puts bang
            puts bang.backtrace
          end
        end

      end

      def setupProject(progress_window)
        begin
          # puts "Load Project Settings:"
          # puts "processing file #{@project_file}"
          # if not loadProjectSettings() then
          #   saveProjectSettings()
          # end

          # puts "Load Session Settings:"
          # puts "processing file #{@session_file}"
          # if not loadSessionSettings() then
          #   saveSessionSettings()
          # end

          importSession(progress_window)

          init_active_modules(progress_window)

          init_passive_modules(progress_window)

          initPlugins(progress_window)

          @settings[:scanlog_dir] = File.expand_path(File.join(@settings[:workspace_path], @settings[:project_name],@settings[:session_name], "scan-logs"))
        rescue => bang
          puts bang
          puts bang.backtrace
        end
      end

      # returns all chats which are in the target scope

      def chatsInScope(chats=nil, scope=nil)
        unique_list = Hash.new
        chatlist = chats.nil? ? @chats : chats
        new_scope = scope.nil? ? @settings[:scope] : scope
        #  puts new_scope.to_yaml
        cis = []
        chat_in_scope = nil
        chatlist.each do |chat|
          next if @settings[:excluded_chats].include?(chat.id)
          uch = uniqueRequestHash(chat.request)

          next if unique_list.has_key?(uch) and @settings[:smart_scan] == true
          unique_list[uch] = nil

          chat_in_scope = chat
          # filter by targets first
          new_scope.each do |s, c_scope|
            chat_in_scope = nil

            if chat.request.site == s then
              chat_in_scope = chat

              if chat_in_scope and c_scope[:root_path] != ''
                chat_in_scope = ( chat.request.path =~ /^(\/)?#{c_scope[:root_path]}/i ) ? chat : nil
              end
              if chat_in_scope and c_scope[:excluded_paths] and c_scope[:excluded_paths].length > 0
                c_scope[:excluded_paths].each do |p|
                  chat_in_scope = nil if ( chat.request.url =~ /#{p}/i )
                end
              end
            end
          end
          #  p chat.request.url if !chat_in_scope.nil?
          cis.push chat_in_scope unless chat_in_scope.nil?
        end
        cis
      end

      def siteInScope?(site)
        #in_scope = false
        @settings[:scope].keys.each do |scope_site|
          return true if scope_site == site
        end
        return false
      end

      def siteSSL?(site)
        @chats.each do |c|
          if c.request.site == site
            return true if c.request.proto =~ /https/
            return false
          end
        end
      end

      #
      # INITIALIZE
      #
      #
      def initialize(project_prefs={})

        @event_dispatcher_listeners = Hash.new
        @settings = nil

        puts "* creating project with following prefs:"
        # puts project_prefs.to_yaml
        setDefaults()

        # reset counters
        Watobo::Core::Chat.new([],[]).resetCounters
        Watobo::Core::Finding.new([],[]).resetCounters

        @project_file = nil
        @session_file = nil

        # UPDATE SETTINGS
        @settings.update(project_prefs)

        @project_file = nil
        if @settings[:project_path] != '' and @settings[:project_name] != ''
          @project_file = File.join(@settings[:project_path], @settings[:project_name] + @settings[:project_settings_file_ext])
        end

        # @:session_settings_file_ext => '.wss',
        #     :project_path => '',
        #     :session_path => '',
        #     :project_name => '',

        t = Time.now
        @date = t.strftime("%m/%d/%Y@%H:%M:%S")

        @interceptor = nil
        @intercept_response = false
        @intercept_request = false
        @unique_requests = Hash.new

        @sessionController = Watobo::Core::Session.new(self)

        @max_response_size = 50000

        #
        @active_checks = []
        @passive_checks = []
        @plugins = []

        # prepare communications
        @chats = []
        @findings = Hash.new

        if @settings[:workspace_path] and @settings[:session_name] and @settings[:project_name]
          project_path = File.join(@settings[:workspace_path], @settings[:project_name])
          session_path = File.join(project_path, @settings[:session_name])
          @session_file = File.join(session_path, @settings[:session_name] + @settings[:session_settings_file_ext])
          @project_file = File.join(project_path, @settings[:project_name] + @settings[:project_settings_file_ext])
          @conversations_path = File.join(session_path, @settings[:conversations_dir])
          @findings_path = File.join(session_path, @settings[:findings_dir])
          @log_path = File.join(session_path, @settings[:logs_dir])
          @scanlog_path = File.join(session_path, "scan-logs")

          [ @conversations_path, @findings_path, @log_path, @scanlog_path ].each do |folder|
            if not File.exists?(folder) then
              puts "create path #{folder}"
              begin
                Dir.mkdir(folder)
              rescue SystemCallError => bang
                puts "!!!ERROR:"
                puts bang
              rescue => bang
                puts "!!!ERROR:"
                puts bang
              end
            end
          end
          # puts "[--- PROJECT PREFS"
          # puts @settings.to_yaml
          # puts "---]"

        end

      end

      #
      # S E S S I O N _ I N I T
      #

      def listSites(prefs={}, &block)
        list = Hash.new

        cprefs = { :in_scope => false,
          :ssl => false
        }
        cprefs.update prefs

        @chats.each do |chat|
          next if list.has_key?(chat.request.site)
          site = chat.request.site
          site = nil if cprefs[:in_scope] == true and not siteInScope?(site)
          site = nil if cprefs[:ssl] and not siteSSL?(site)
          unless site.nil?
            yield site if block_given?
            list[site] = nil
          end
          #  yield chat.request.site if chat.request.site
        end
        return list.keys
      end

      def listDirs(site, list_opts={}, &block)
        opts = { :base_dir => "",
          :include_subdirs => true
        }
        opts.update(list_opts) if list_opts.is_a? Hash
        list = Hash.new
        @chats.each do |chat|
          next if chat.request.site != site
          next if list.has_key?(chat.request.path)
          next if opts[:base_dir] != "" and chat.request.path !~ /^#{Regexp.quote(opts[:base_dir])}/
          subdirs = chat.request.subDirs
          subdirs.each do |dir|
            next if dir.nil?
            next if list.has_key?(dir)
            list[dir] = :path
            if opts[:include_subdirs] == true then
              yield dir if block
            else
              d = dir.gsub(/#{Regexp.quote(opts[:base_dir])}/,"")
              yield dir unless d =~ /\// and block
              # otherwise it is a subdir of base_dir
            end
          end
        end
      end

      private

      def listChatIds(path, pattern)
        id_list = []
        Dir.foreach(path) do |file|
          if file =~ /#{pattern}/ then
            id = file.gsub!(/-#{pattern}/,'').to_i
            id_list.push id
          end
        end
        return id_list.uniq.sort
      end

      def importSession(progress_window = nil)
        puts "* Import Session:"
        puts "+ Conversation Path:\n>> #{File.expand_path(@conversations_path)}"

        puts
        chatIds = listChatIds(@conversations_path, "chat")
        findingIds = listChatIds(@findings_path, "finding")

        numChats = chatIds.length
        numFindings = findingIds.length
        numImports = numChats + numFindings
        pc = 0

        progress_window.updateProgress(:total => numImports, :job => "Import Conversation")
        begin
          chatIds.each do |id|
            progress_window.updateProgress(:progress => 1) if progress_window

            fname = File.join(@conversations_path, "#{id}-chat")
            chat = Watobo::Utils.loadChatYAML(fname)
            addChat(chat, :run_passive_checks => false) if chat
          end
        rescue => bang
          puts "!!!ERROR: Could not import conversations"
          puts bang
          puts bang.backtrace
        end

        puts "+ Findings Path:\n>> #{File.expand_path(@findings_path)}"
        progress_window.updateProgress(:job => "Import Findings") if progress_window

        begin
          findingIds.each do |id|

            #  @settings[:progress_window].progress(pc, :total => numImports, :job_name => "Import Findings", :step_name => "load finding: #{id}") if @settings[:progress_window]
            progress_window.updateProgress(:progress => 1) if progress_window
            #  request, response = Watobo::Utils.loadChat(@findings_path, id.to_s, "-request", "-response")
            fname = File.join(@findings_path, "#{id}-finding")
            finding = Watobo::Utils.loadFindingYAML(fname)
            #  puts finding
            addFinding(finding) if finding

          end
        rescue => bang
          puts "!!!ERROR: Could not import finding [#{id}]"
          puts bang

        end

        #pw.close
      end

      def setDefaults()
        @settings = {
          :excluded_chats => [],
          :scope => Hash.new,
          #:project_prefs => {
          :project_settings_file_ext => '.wps',
          :session_settings_file_ext => '.wss',
          :project_path => '',
          :session_path => '',
          :project_name => '',
          :session_name => '',
          :module_dir => 'modules',
          :plugin_dir => 'plugins',
          :conversations_dir => "conversations",
          :findings_dir => "findings",
          :logs_dir => "logs",
          #},

          #:scan_prefs => {
          :custom_error_patterns => [],
          :max_parallel_checks => 15,
          :excluded_parms => [ "__VIEWSTATE"],
          :non_unique_parms => [],
          :smart_scan => true,
          :run_passive_checks => false,
          #},
          :policy => { :name => 'Default',
          :list => {
          'Default' => {
          'Watobo::Modules::Active::Directories::Dir_indexing' => false,
          'Watobo::Modules::Active::Discovery::Http_methods' => false,
          'Watobo::Modules::Active::Domino::Domino_db' => false,
          'Watobo::Modules::Active::Sap::Its_commands' => false,
          'Watobo::Modules::Active::Sap::Its_services' => false,
          'Watobo::Modules::Active::Sap::Its_service_parameter' => false,
          'Watobo::Modules::Active::Sap::Its_xss' => false,
          'Watobo::Modules::Active::Sqlinjection::Sqli_simple' => true,
          'Watobo::Modules::Active::Sqlinjection::Sql_boolean' => true,
          'Watobo::Modules::Active::Sqlinjection::Sql_numerical' => false,
          'Watobo::Modules::Active::Xss::Xss_simple' => true,
          }
          }
          },

          #@interface = Hash.new
          #:session_management => {
          :csrf_patterns => [
          "name=\"(token)\" value=\"([0-9a-zA-Z!-]*)\"",
          "(token)=([-0-9a-zA-Z_:]*)(;|&)?"
          ],
          :csrf_request_ids => Hash.new,
          
          :login_chat_ids => [],
          :sid_patterns => [
          "name=\"(sessid)\" value=\"([0-9a-zA-Z!-]*)\"",
          "(sessid)=([-0-9a-zA-Z_:]*)(;|&)?",
          '(SESSIONID)=([-0-9a-zA-Z_:\.\(\)]*)(;|&)?',
          "(PHPSESSID)=([0-9a-zA-Z]*)(;|&)?",
          '(ASPSESSIONID)\w*=([0-9a-zA-Z]*)(;|&)?',
          "(MYSAPSSO2)=([0-9a-zA-Z.=%]*)(;|&)?",
          "(ELEXIRSID)=([0-9a-zA-Z!-]*)(;|&)?",
          "(SLSID)=([0-9a-zA-Z!-]*)(;|&)?",
          "(sid)=([0-9a-z]*)(')?", #servlet?sid=9912ad967cc578a12ada85d91f841e18')
          '(saplb_\*)=([-0-9a-zA-Z_:\(\)]*)(;|&)?',
          "(DomAuthSessId)=([0-9a-zA-Z]*)(;|&)?",
          '(wgate)\/([\w]{4,}\/[\w=~]*)(;|&|\'|")?', # SAP ITS WGATE Session Handling
          '(session)=([-0-9a-zA-Z_:\.]*)(;|&)?'       # SAP ITS Session Cookie
          ],
          :logout_signatures => [
          "^Location.*login"
          ],
          #},

          #      :settings => {
          #        :module_path => ''
          #      },
          :history => [],
          :www_auth => {}
        }
      end

      def initPlugins(progress_window=nil)
        puts
        puts "=== Initialize Plugins ==="
        ignore = ['.','..','.svn']
        # ignore_modules = ["http_methods.rb"]
        ignore_modules = []
        begin
          plugin_path = File.join($watobo_base, @settings[:plugin_dir])

          Dir.chdir(plugin_path)
          puts "Plugins Directory: \"#{Dir.getwd}\""
        rescue StandardError => bang
          puts bang
          puts "\nError changing into source directory \"#{plugin_path}\"!"
          exit
        end

        progress_window.updateProgress(:total => Dir.entries(".").length-2, :title => "Loading Plugins") if progress_window
        Dir.foreach(".") do |group|
          if File.ftype(group) == "directory" and !ignore.include?(group)
            Dir.foreach("./#{group}") do |modules|
              if File.ftype("./#{group}/#{modules}") == "file" and modules =~ /.rb/ and not ignore_modules.include?(modules)
                begin
                  plugin_file = File.join(plugin_path, group, modules)

                  notify(:logger, "loading plugin: #{plugin_file}")
                  # p plugin_file
                  #require "#{@settings[:plugin_path]}/#{group}/#{modules}"
                  require plugin_file
                  # load "#{@settings[:module_path]}/#{modules}/#{check}"
                  group_class = group.slice(0..0).upcase + group.slice(1..-1).downcase
                  #
                  module_class = modules.slice(0..0).upcase + modules.slice(1..-1).downcase
                  module_class.sub!(".rb","")
                  # How to get a class constant out of a string ??? Here we go ...
                  #puts "* Init Plugin Watobo::Plugin::#{module_class}"
                  class_constant = Watobo.class_eval("Watobo::Plugin::#{group_class}::#{module_class}")

                  # IMPORTANT: Change to plugin directory first before creating a class instance!!!
                  Dir.chdir(group)
                  puts "+ #{module_class}"
                  new_plugin = class_constant.new(FXApp.instance, self)
                  # new_plugin.subscribe(:new_finding){ |finding|
                  # addFinding(finding)
                  # }
                  @plugins.push new_plugin
                  progress_window.updateProgress(:progress => 1) if progress_window
                  # @interface.log("#{module_class} loaded.")
                  notify(:logger, LOG_INFO,"#{module_class} loaded.")
                  Dir.chdir("..")
                rescue => bang
                  puts bang
                  notify(:logger, LOG_INFO, "problems loading plugin: #{plugin_file}")
                end
              end
            end
          end

        end
        puts "-----------------"
        puts "Loaded #{@plugins.length} Plugins."
      end

      def init_active_modules(progress_window = nil)
        puts
        puts "=== Initialize Active Modules ==="
        ignore = ['.','..','.svn']
        # ignore_modules = ["http_methods.rb"]
        ignore_modules = []

        active_path = File.join($watobo_base, @settings[:module_dir],"active")
        begin
          Dir.chdir(active_path)
          puts "Active Modules Directory: \"#{Dir.getwd}\""
        rescue StandardError => bang
          puts bang
          puts "\nError changing into source directory \"#{active_path}\"!"
          exit
        end

        Dir.foreach(".") do |group|

          if File.ftype(group) == "directory" and !ignore.include?(group)

            Dir.foreach("./#{group}") do |modules|
              if File.ftype("./#{group}/#{modules}") == "file" and modules =~ /.rb$/
                begin
                  module_file = File.join(active_path, group, modules)
                  notify(:logger, LOG_DEBUG, "loading module: #{module_file}")
                  #require "#{active_path}/#{group}/#{modules}"
                  require module_file
                  # load "#{@settings[:module_path]}/#{modules}/#{check}"
                  group_class = group.slice(0..0).upcase + group.slice(1..-1).downcase
                  #
                  module_class =modules.slice(0..0).upcase + modules.slice(1..-1).downcase
                  module_class.sub!(".rb","")
                  # How to get a class constant out of a string ??? Here we go ...

                  class_constant = Watobo.class_eval("Watobo::Modules::Active::#{group_class}::#{module_class}")

                  @active_checks.push class_constant.new(self)
                  puts "+ #{module_class}"
                  notify(:logger, LOG_INFO, "#{module_class} loaded.")

                rescue => bang
                  puts bang
                  notify(:logger, LOG_DEBUG, "problems loading module: #{@settings[:module_path]}/active/#{group}/#{modules}")
                end

              end
            end
          end

        end

        puts "-----------------"
        puts "Loaded #{@active_checks.length} Active Modules."

      end

      def applyPreferences(prefs)

      end

      def init_passive_modules(progress_window = nil)
        puts
        puts "=== Initialize Passive Modules ==="
        ignore = ['.','..','.svn']
        # puts "get passive modules from path #{@settings[:module_path]}/passive"

        passive_path = File.join($watobo_base, @settings[:module_dir],"passive")
        begin
          Dir.chdir(passive_path)
          puts "Passive Modules Directory: \"#{Dir.getwd}\""
        rescue StandardError => bang
          puts bang
          puts "\nError changing into source directory \"#{passive_path}\"!"
          exit
        end
        Dir.foreach(".") do |modules|
          begin
            if modules =~ /.rb$/
              module_file = File.join(passive_path, modules)
              # puts "#{@settings[:module_path]}/passive/#{modules}"
              require module_file
              #  puts "+ #{modules}"
              # load "#{@settings[:module_path]}/#{modules}/#{check}"
              classname = modules.slice(0..0).upcase + modules.slice(1..-1).downcase
              classname.sub!(".rb","")

              # How to get a class constant out of a string ??? Here we go ...

              class_constant = Watobo.class_eval("Watobo::Modules::Passive::#{classname}")

              @passive_checks.push class_constant.new(self)
              puts "+ #{classname}"
              notify(:logger, LOG_DEBUG, "#{modules} loaded.")
            end
          rescue => bang
            puts "!!!"
            puts bang
          end
        end
        puts "-----------------"
        puts "Loaded #{@passive_checks.length} Passive Modules."
      end

      private

    end

  end # Framework
end # Watobo