# .
# session.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
require 'socket'
require 'timeout'
require 'openssl'
require 'optparse'
require 'digest/md5'
require 'lib/httpparser.rb'
require 'lib/qShapers.rb'
require 'lib/http_socket.rb'
require 'lib/external/ntlm/ntlm'

module Watobo
  module Core
    class Proxy
      attr :login
      attr :name
      attr :host
      attr :port
      attr :login
      def setCredentials(creds)
        @login = Hash.new
        @login.update creds
      end

      def has_login?
        return false if @login.nil?
        return true
      end

      def initialize(prefs)
        @login = nil
        @name = prefs[:name]
        @host = prefs[:host]
        @port = prefs[:port]

      end
    end

    class Session

      @@settings = Hash.new
      @@proxy = Hash.new

      @@session_lock = Mutex.new
      @@csrf_lock = Mutex.new

      @@login_mutex = Mutex.new
      @@login_cv = ConditionVariable.new
      def subscribe(event, &callback)
        (@event_dispatcher_listeners[event] ||= []) << callback
      end

      def clearEvents(event)
        @event_dispatcher_listener[event].clear
      end

      def notify(event, *args)
        if @event_dispatcher_listeners[event]
          @event_dispatcher_listeners[event].each do |m|
            m.call(*args) if m.respond_to? :call
          end
        end
      end

      def runLogin(chat_list, prefs={})
        login_prefs = Hash.new
        login_prefs.update prefs
        dummy = {:ignore_logout => true, :update_sids => true, :update_session => true, :update_contentlength => true}
        login_prefs.update dummy
        # puts "* RUN LOGIN"
        #  puts login_prefs.to_yaml
        chat_list.each do |chat|
          test_req = chat.copyRequest
          request, response = doRequest(test_req, login_prefs)
        end
        @@login_cv.signal
        # exit
        #  print "L]"
      end

      #     def sessionSettings=(prefs)
      #       applySessionSettings(prefs)
      #     end

      def sessionSettings()
        @@settings
      end

      def sendHTTPRequest(request, prefs={})

        begin
          @lasterror = nil
          response_header = nil
          site = request.site
          proxy = getProxy(site)

          unless proxy.nil?
            host = proxy.host
            port = proxy.port
          else
            host = request.host
            port = request.port
          end
          # check if hostname is valid and can be resolved
          hostip = IPSocket.getaddress(host)

          # update current preferences, prefs given here are stronger then global settings!
          current_prefs = Hash.new
          [:update_session, :update_sids, :update_contentlength, :ssl_cipher, :www_auth].each do |k|
            current_prefs[k] = prefs[k].nil? ? @session[k] : prefs[k]
          end

          updateSession(request) if current_prefs[:update_session] == true

          #---------------------------------------
          request.removeHeader("^Proxy-Connection") #if not use_proxy
          request.removeHeader("^Connection") #if not use_proxy
          request.removeHeader("^Accept-Encoding")
          # If-Modified-Since: Tue, 28 Oct 2008 11:06:43 GMT
          # If-None-Match: W/"3975-1225192003000"
          request.removeHeader("^If-")
          #  puts
          #  request.each do |line|
          #  puts line.unpack("H*")
          #end
          #puts
          if current_prefs[:update_contentlength] == true then
            request.fix_content_length()
          end

          # puts "========== Add Headers"
          request.addHeader("Connection", "Close") #if not use_proxy
          request.addHeader("Proxy-Connection", "Close") #if not use_proxy
          request.addHeader("Accept-Encoding", "None") #don't want encoding
          #request.add_header("Via", "Watobo") if use_proxy
          #puts request
          # puts "=============="
        rescue SocketError
          puts "!!! unknown hostname #{host}"
          return nil, "WATOBO: Could not resolve hostname #{host}", nil
        rescue => bang
          puts bang
          puts bang.backtrace
        end

        begin
          unless proxy.nil?
            # connection requires proxy
            puts "* use proxy #{proxy.name}"

            if request.is_ssl? then
              tcp_socket = sslProxyConnect(request, proxy)
              socket = sslConnect(tcp_socket)

            else
              # check for regular proxy authentication
              socket = doProxyRequest(request, proxy)
            end
          else
            # direct connection to host
            tcp_socket = nil
            #  timeout(6) do
#puts "* no proxy - direct connection"
            tcp_socket = TCPSocket.new( host, port )
            tcp_socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
            tcp_socket.sync = true

            socket = request.is_ssl? ? sslConnect(tcp_socket) : tcp_socket
#puts socket.class
            # remove URI before sending request but cache it for restoring request
            uri_cache = nil
            uri_cache = request.removeURI #if proxy.nil?
            #
            # WWW Authentication
          #  puts current_prefs[:www_auth].to_yaml

            if current_prefs[:www_auth].has_key?(site)
              case current_prefs[:www_auth][site][:type]
              when AUTH_TYPE_NTLM
               # puts "* found NTLM credentials for site #{site}"
                socket, response_header = doNtlmAuth(socket, request, current_prefs[:www_auth][site])
                request.restoreURI(uri_cache)
                response_header.extend Watobo::Parser::Url
                response_header.extend Watobo::Parser::Web10
                return socket, request, response_header
              else
                puts "* Unknown Authentication Type: #{current_prefs[:www_auth][site][:type]}"
              end
            else

              # puts request
              data = request.join + "\r\n"

              socket.print data
              response_header = readHTTPHeader(socket, current_prefs)

              # RESTORE URI FOR HISTORY/LOG
              request.restoreURI(uri_cache) #if proxy.nil?
              #  puts request.first
            end
          end
        rescue Errno::ECONNREFUSED
          # response.push "WATOBO: connection refused (#{host}:#{port})"
          print "C!"
          request = "WATOBO: connection refused (#{host}:#{port})"
          socket = nil
        rescue Errno::ECONNRESET
          puts '!!! connection reset'
        rescue Errno::EHOSTUNREACH
          request = "WATOBO: host unreachable (#{host}:#{port})"
          socket = nil
        rescue Timeout::Error
          request = "WATOBO: TimeOut (#{host}:#{port})\n"
          socket = nil
        rescue Errno::ETIMEDOUT
          #response.push
          #response.push request.first if request
          request = "WATOBO: TimeOut (#{host}:#{port})"
          socket = nil
          #rescue EOFError
          #puts "EOF"
          #printf("O\r\n")
        rescue Errno::ENOTCONN
          puts "!!!ENOTCONN"
        rescue OpenSSL::SSL::SSLError
          request = nil
          socket = nil
          #  puts "!!! SSL-Error"
          # print "E"
        rescue => bang
          puts "!!! Uuuuppppsssss ...."
          puts request.to_yaml
          request = nil
          socket = nil

          puts bang

          puts bang.backtrace
        end

        # @project.extendRequest(request) if request

        return socket, request, response_header
      end

      #
      # getHTTPHeader
      #
      def getHTTPHeader_UNUSED(raw_request, prefs)
        begin
          #  print(".")
          socket, sent_request = sendHTTPRequest(raw_request, prefs)
          if socket.nil? then
            #   print("x")
            return nil, raw_request, sent_request
          end
          header = readHTTPHeader(socket)
          #print("o")
        rescue => bang
          puts bang
        end

        return socket, sent_request, header
      end

      # +++ doRequest(request)  +++
      # + function:
      #
      def doRequest(request, opts={} )
        begin
          @session.update opts

          # puts "#[#{self.class}]" + @session[:csrf_requests].first.object_id.to_s
          unless @session[:csrf_requests].empty? or @session[:csrf_patterns].empty?
            csrf_cache = Hash.new
            @session[:csrf_requests].each do |req|
              copy = YAML.load(YAML.dump(req))
              copy.map!{|l|
                x = l.strip + "\r\n"
                x = l if l == copy.last
                x
              }
              # now extend the new request with the Watobo mixins
              copy.extend Watobo::Parser::Url
              copy.extend Watobo::Parser::Web10
              copy.extend Watobo::Framework::Shaper::Web10

              # p copy.first

              updateCSRFToken(csrf_cache, copy)
              socket, csrf_request, csrf_response = sendHTTPRequest(copy, opts)

              next if socket.nil?
              #  p "*"
              #    csrf_response = readHTTPHeader(socket)
              readHTTPBody(socket, csrf_response, csrf_request, opts)

              next if csrf_response.body.nil?
              updateCSRFCache(csrf_cache, csrf_request, csrf_response.body) if csrf_response.content_type =~ /text\//

              socket.close
            end
            #p @session[:csrf_requests].length
            updateCSRFToken(csrf_cache, request)
          end

          socket, request, response = sendHTTPRequest(request, opts)

          if socket.nil?
            return nil, request
          end
        
          update_sids(request.site, response.headers) if @session[:update_sids] == true

          readHTTPBody(socket, response, request, opts)

          unless response.body.nil?
            update_sids(request.site, response.body) if @session[:update_sids] == true and response.content_type =~ /text\//
          end

          socket.close

        rescue  => bang
          puts "! Error in doRequest"
          puts "Exception!! in Module #{Module.nesting[0].name}"
          puts bang
          puts bang.backtrace
          @lasterror = bang
          # raise
          # ensure
        end

        response.extend Watobo::Parser::Web10
        return request, response
      end

      def addProxy(prefs=nil)
        return if prefs.nil?
        puts "* add proxy"
        puts prefs.to_yaml
        proxy = Proxy.new(:name => prefs[:name], :host => prefs[:host], :port => prefs[:port])

        proxy.setCredentials(prefs[:credentials]) unless prefs[:credentials].nil?

        unless prefs[:site].nil?
          @@proxy[prefs[:site]] = proxy
        else
          @@proxy[:default] = proxy
        end
      end

      def getProxy(site=nil)
        unless site.nil?
          return @@proxy[site] unless @@proxy[site].nil?
        end
        return @@proxy[:default]
      end

      #
      # INITIALIZE
      #
      # Possible preferences:
      # :proxy => '127.0.0.1:port'
      # :valid_sids => Hash.new,
      # :sid_patterns => [],
      # :logout_signatures => [],
      # :update_valid_sids => false,
      # :update_sids => false,
      # :update_contentlength => true
      def initialize(session_id, prefs={})
        @event_dispatcher_listeners = Hash.new
        @session = nil

        session_id = Digest::MD5.hexdigest(Time.now.to_f.to_s) if session_id.nil?

        if !@@settings[session_id] then
          @@settings[session_id] = {
            :valid_sids => Hash.new,
            :sid_patterns => [],
           # :valid_csrf_tokens => Hash.new,
            :csrf_patterns => [],
            :csrf_requests => [],
            :logout_signatures => [],
            :logout_content_types => Hash.new,
            :update_valid_sids => false,
            :update_sids => false,
            :update_session => true,
            :update_contentlength => true,
            :login_chats => [],
            :www_auth => Hash.new,
            :proxy_auth => Hash.new
          }
        end

        @session = @@settings[session_id] # shortcut to settings
        @session.update prefs

      #  @valid_csrf_tokens = Hash.new

        addProxy( prefs[:proxy] ) if prefs.is_a? Hash and prefs[:proxy]

        @socket = nil

        # TODO: Implement switches for URL-Encoding (http://www.blooberry.com/indexdot/html/topics/urlencoding.htm)
        # TODO: Implement switches for Following Redirects
        # TODO: Implement switches for Logging, Debugging, ...
      end

      def readHTTPBody(socket, response, request, prefs={})
        clen = response.content_length
        data = ""
        #   timeout(5) do
        begin
          if  clen > 0
          #  puts "* read #{clen} bytes for body"
            Watobo::Framework::HTTP.read_body(socket, :max_bytes => clen) { |c|
              data += c
              break if data.length == clen
            }
          elsif response.is_chunked?
            Watobo::Framework::HTTP.readChunkedBody(socket) { |c|
              data += c
            }
          else
           # puts "* no content-length information ... mmmmmpf"
            eofcount = 0
            Watobo::Framework::HTTP.read_body(socket) do |c|
              data += c
            end

          end
        rescue => e
          puts "!!! ERROR: read_response"
          puts e
          puts e.backtrace
        end
        # end

        response.push data
        unless prefs[:ignore_logout]==true  or @session[:logout_signatures].empty?
          notify(:logout, self) if loggedOut?(response)
        end

        update_sids(request.site, response) if prefs[:update_sids] == true

      end

      private

      def doNtlmAuth(socket, request, ntlm_credentials)
        response_header = nil
        begin
          auth_request = Watobo::Utils::copyObject(request)
          auth_request.extend Watobo::Parser::Url
          auth_request.extend Watobo::Parser::Web10
          auth_request.extend Watobo::Framework::Shaper::Web10

          ntlm_challenge = nil
          t1 = Net::NTLM::Message::Type1.new()
          msg = "NTLM " + t1.encode64

          auth_request.removeHeader("Connection")
          auth_request.removeHeader("Authorization")

          auth_request.addHeader("Authorization", msg)
          auth_request.addHeader("Connection", "Keep-Alive")

#          puts "============= T1 ======================="
          #    puts auth_request
          data = auth_request.join + "\r\n"
#puts "= REQUEST ="

#auth_request.each do |l|
#  puts l
#  puts l.unpack("H*")
#end
#puts "-----------"
          socket.print data
          #  puts "-----------------"
          response_header = []
          while (line = socket.gets)
 #              puts line
            response_header.push line
            if line =~ /^HTTP\/\d\.\d (\d+) (.*)/ then
              rcode = $1.to_i
              rmsg = $2
            end
            if line =~ /^WWW-Authenticate: (NTLM) (.+)\r\n/
              ntlm_challenge = $2
            end
            if line =~ /^Content-Length: (\d{1,})\r\n/
              clen = $1.to_i
            end
            break if line.strip.empty?
          end
  #        puts "==================="
                                    
          #if rcode == 200 # Ok
           # puts "* seems request doesn't need authentication"
          #  return socket, response_header
          if rcode == 401 #Authentication Required
           # puts "* got ntlm challenge: #{ntlm_challenge}"
            return socket, response_header if ntlm_challenge.nil?
          else
           # puts "! arrgh .... :("
           # puts response_header
            return socket, response_header
          end

          # reading rest of response
          Watobo::Framework::HTTP.read_body(socket, :max_bytes => clen){ |d| }
                  
          t2 = Net::NTLM::Message.decode64(ntlm_challenge)
          t3 = t2.response({:user => ntlm_credentials[:username],
            :password => ntlm_credentials[:password],
              :domain => ntlm_credentials[:domain]}, 
              {:workstation => ntlm_credentials[:workstation], :ntlmv2 => true})
        
        
   #     puts "* NTLM-Credentials: #{ntlm_credentials[:username]},#{ntlm_credentials[:password]}, #{ntlm_credentials[:domain]}, #{ntlm_credentials[:workstation]}"
          auth_request.removeHeader("Authorization")
          auth_request.removeHeader("Connection")

          auth_request.addHeader("Connection", "Close")

          msg = "NTLM " + t3.encode64
          auth_request.addHeader("Authorization", msg)
    #      puts "============= T3 ======================="
           
          data = auth_request.join + "\r\n"

  #        puts data
          socket.print data

          response_header = []
          while (line = socket.gets)
   #            puts line
            response_header.push line
            if line =~ /^HTTP\/\d\.\d (\d+) (.*)/ then
              rcode = $1.to_i
              rmsg = $2
            end
            break if line.strip.empty?
          end

          if rcode == 200 # Ok
            puts "* authentication successfull [OK]"
            return socket, response_header
          elsif rcode == 401 # Authentication Required
            # TODO: authorization didn't work -> do some notification
            # ...
            puts "* could not authenticate"
          end

          return socket, response_header
        rescue => bang
          puts "!!! ERROR: in ntlm_auth"
          puts bang
          
        puts bang.backtrace
          return nil, nil
        end
      end

def sslConnect(tcp_socket, current_prefs = {} )
        begin
        @ctx = OpenSSL::SSL::SSLContext.new()
        @ctx.key = nil
        @ctx.cert = nil

        @ctx.ciphers = current_prefs[:ssl_cipher] if current_prefs[:ssl_cipher]

        @ctx.tmp_dh_callback = proc { |*args|
          OpenSSL::PKey::DH.new(
          OpenSSL::PKey::DH.new(128)
          )
        }

        socket = OpenSSL::SSL::SSLSocket.new(tcp_socket, @ctx)

        socket.connect
        socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
        return socket
        rescue => bang
          puts "!sslConnect"
          puts bang
          puts bang.backtrace
        end
      end

      def sslProxyConnect(request, proxy)
        begin
          tcp_socket = nil
          auth_request = Watobo::Utils::copyObject(request)
          auth_request.extend Watobo::Parser::Url
          auth_request.extend Watobo::Parser::Web10
          auth_request.extend Watobo::Framework::Shaper::Web10
          #  timeout(6) do

          tcp_socket = TCPSocket.new( proxy.host, proxy.port)
          tcp_socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
          tcp_socket.sync = true
          #  end

          if request.is_ssl? then
            auth_request.shift
            auth_request.unshift "CONNECT #{request.host}:#{request.port} HTTP/1.0\r\n"
          end

          return nil unless proxy.login[:type] == AUTH_TYPE_NTLM

          ntlm_challenge = nil
          t1 = Net::NTLM::Message::Type1.new()
          msg = "NTLM " + t1.encode64

          auth_request.removeHeader("Proxy-Connection")
          auth_request.removeHeader("Connection")

          auth_request.addHeader("Proxy-Authorization", msg)
          auth_request.addHeader("Proxy-Connection", "Keep-Alive")
          auth_request.addHeader("Pragma", "no-cache")

          puts "============= T1 ======================="
          puts auth_request
          data = auth_request.join + "\r\n"

          tcp_socket.print data
          puts "-----------------"
          while (line = tcp_socket.gets)
            puts line
            if line =~ /^HTTP\/\d\.\d (\d+) (.*)/ then
              rcode = $1.to_i
              rmsg = $2
            end
            if line =~ /^Proxy-Authenticate: (NTLM) (.+)\r\n/
              ntlm_challenge = $2
            end
            break if line.strip.empty?
          end

          Watobo::Framework::HTTP.read_body(tcp_socket) { |d|
            puts d
          }

          if rcode == 200 # Ok
            puts "* check proxy connection [OK]"
          elsif rcode == 407 # ProxyAuthentication Required
            return nil if ntlm_challenge.nil?
          end

          t2 = Net::NTLM::Message.decode64(ntlm_challenge)
          t3 = t2.response( { :user => proxy.login[:username], 
            :password => proxy.login[:password], 
            :domain => proxy.login[:domain] }, 
             { :workstation => proxy.login[:workstation], :ntlmv2 => true } )
          auth_request.removeHeader("Proxy-Authorization")
          
          msg = "NTLM " + t3.encode64
          auth_request.addHeader("Proxy-Authorization", msg)
          puts "============= T3 ======================="
          puts auth_request
          data = auth_request.join + "\r\n"

          tcp_socket.print data
          puts "-----------------"

          while ( line = tcp_socket.gets )
            if line =~ /^HTTP\/\d\.\d (\d+) (.*)/ then
              rcode = $1.to_i
              rmsg = $2.strip
            end

            break if line.length < 3
          end
          if rcode == 200 # Ok
            puts "* authentication successfull"
          elsif rcode == 407 # ProxyAuthentication Required
            # if rcode is still 407 authentication didn't work -> break
            closeSocket(tcp_socket)

            return nil
          else
            puts "! check proxy connection [FALSE]"
            puts ">  #{rcode} #{rmsg} <"
          end

          return tcp_socket
        rescue => bang
          puts bang
          puts bang.backtrace

        end
        return nil
      end

      def doProxyRequest(request, proxy)

        begin
          tcp_socket = nil
          auth_request = Watobo::Utils::copyObject(request)
          auth_request.extend Watobo::Parser::Url
          auth_request.extend Watobo::Parser::Web10
          auth_request.extend Watobo::Framework::Shaper::Web10
          #  timeout(6) do

          tcp_socket = TCPSocket.new( proxy.host, proxy.port)
          tcp_socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
          tcp_socket.sync = true
          #  end

          return nil unless proxy.login[:type] == AUTH_TYPE_NTLM

          ntlm_challenge = nil
          t1 = Net::NTLM::Message::Type1.new()
          msg = "NTLM " + t1.encode64

          auth_request.removeHeader("Proxy-Connection")
          auth_request.removeHeader("Connection")

          auth_request.addHeader("Proxy-Authorization", msg)
          auth_request.addHeader("Proxy-Connection", "Keep-Alive")
          auth_request.addHeader("Pragma", "no-cache")

          #   puts "============= T1 ======================="
          #    puts auth_request
          data = auth_request.join + "\r\n"

          tcp_socket.print data
          #  puts "-----------------"
          while (line = tcp_socket.gets)
            #   puts line
            if line =~ /^HTTP\/\d\.\d (\d+) (.*)/ then
              rcode = $1.to_i
              rmsg = $2
            end
            if line =~ /^Proxy-Authenticate: (NTLM) (.+)\r\n/
              ntlm_challenge = $2
            end
            if line =~ /^Content-Length: (\d{1,})\r\n/
              clen = $1.to_i
            end
            break if line.strip.empty?
          end

          #puts "* reading #{clen} bytes"
          Watobo::Framework::HTTP.read_body(tcp_socket, :max_bytes => clen){ |d|
            #puts d
          }

          if rcode == 200 # Ok
            puts "* check proxy connection [OK]"
          elsif rcode == 407 # ProxyAuthentication Required
            return nil if ntlm_challenge.nil?
          end

          t2 = Net::NTLM::Message.decode64(ntlm_challenge)
          t3 = t2.response({:user => proxy.login[:username], :password => proxy.login[:password], :workstation => proxy.login[:workstation], :domain => proxy.login[:domain]}, {:ntlmv2 => true})
          auth_request.removeHeader("Proxy-Authorization")
          auth_request.removeHeader("Proxy-Connection")

          auth_request.addHeader("Proxy-Connection", "Close")
          auth_request.addHeader("Pragma", "no-cache")
          msg = "NTLM " + t3.encode64
          auth_request.addHeader("Proxy-Authorization", msg)
          # puts "============= T3 ======================="
          # puts auth_request
          data = auth_request.join + "\r\n"

          tcp_socket.print data

          return tcp_socket
        rescue => bang
          puts bang
          puts bang.backtrace

        end
        return nil
      end

      def loggedOut?(response, prefs={})
        begin
          return false if @session[:logout_signatures].empty?
          response.each do |line|
            @session[:logout_signatures].each do |p|
              #     puts "!!!*LOGOUT*!!!" if line =~ /#{p}/
              return true if line =~ /#{p}/
            end
          end
        rescue => bang
          puts bang
          puts bang.backtrace
        end
        return false
      end

      def readHTTPHeader(socket, prefs={})
        header = []
        begin
          Watobo::Framework::HTTP.read_header(socket) do |line|
            # puts line.unpack("H*")
            header.push line
          end
        rescue => bang
          puts "!ERROR: read_header"
          return nil
        end

        header.extend Watobo::Parser::Url
        header.extend Watobo::Parser::Web10
        #  update_sids(header)

        #  update_sids(request.site, response) if prefs[:update_sids] == true

        #  unless prefs[:ignore_logout]==true or @session[:logout_signatures].empty?
        #    notify(:logout, self) if loggedOut?(header)
        #  end

        return header
      end

      #     def read_response(socket)

      #       return response
      #    end

      def update_sids(site, response)
        #   p "* update sids"
        Thread.new{
          begin
            #site = request.site
            @@session_lock.synchronize do
              response.each do |line|
                # puts line
                @session[:sid_patterns].each do |pat|
                  if line =~ /#{pat}/i then
                    sid_key = Regexp.quote($1.upcase)
                    sid_value = $2
                    #print "U"
                    #   puts "GOT NEW SID (#{sid_key}): #{sid_value}"
                    @session[:valid_sids][site] = Hash.new if @session[:valid_sids][site].nil?
                    @session[:valid_sids][site][sid_key] = sid_value
                  end
                end

              end
            end
          rescue => bang
            puts bang
            puts bang.backtrace

          end
        }
      end

      def updateCSRFCache(csrf_cache, request, response)
       # puts "=UPDATE CSRF CACHE"
        # Thread.new{
        begin
          site = request.site
          @@csrf_lock.synchronize do
            response.each do |line|
              # puts line
              @session[:csrf_patterns].each do |pat|
                if line =~ /#{pat}/i then
                  token_key = Regexp.quote($1.upcase)
                  token_value = $2
                  #print "U"
                #  puts "GOT NEW TOKEN (#{token_key}): #{token_value}"
                  #   @session[:valid_csrf_tokens][site] = Hash.new if @session[:valid_csrf_tokens][site].nil?
                  #   @session[:valid_csrf_tokens][site][token_key] = token_value
                  csrf_cache[token_key] = token_value
                end
              end

            end
          end
        rescue => bang
          puts bang
          puts bang.backtrace

        end
        # }
      end

      def closeSocket(socket)
        #puts socket.class
        begin
          #puts socket.class
          #if socket.class.to_s =~ /SSLSocket/
          if socket.is_a? OpenSSL::SSL::SSLSocket
            socket.io.shutdown(2)
          else
            socket.shutdown(2)
          end
          socket.close
        rescue => bang
          puts bang
          puts bang.backtrace
        end
      end

      def updateSessionSettings(settings={})
        [:valid_sids, :sid_patterns, :logout_signatures, :logout_content_types, :update_valid_sids, :update_sids, :update_session, :update_contentlength, :login_chats].each do |k|
          @session[k] = settings[k] if settings[k]
        end
      end

      def updateSession(request)
        @@session_lock.synchronize do
          if @session[:valid_sids].has_key?(request.site)
            # puts "* found sid for site: #{request.site}"
            request.map!{ |line|
              res = line
              @session[:sid_patterns].each do |pat|
                begin
                  if line =~ /#{pat}/i then
                    next if $~.length < 3
                    sid_key = Regexp.quote($1.upcase)
                    old_value = $2

                    if @session[:valid_sids][request.site].has_key?(sid_key) then
                      if not old_value =~ /#{@session[:valid_sids][request.site][sid_key]}/ then # sid value has changed and needs update
                        #      print "S"
                        #    puts "+ update sid #{sid_key}"
                        #    puts "-OLD: #{old_value}"
                        #    puts "-NEW: #{@session[:valid_sids][request.site][sid_key]}"

                        #      puts "---"
                        # dummy = Regexp.quote(old_value)
                        res = line.gsub!(/#{Regexp.quote(old_value)}/, @session[:valid_sids][request.site][sid_key])
                        if not res then puts "!!!could not update sid (#{sid_key})"; end
                        #     puts "->#{line}"

                      end
                    end
                  end
                rescue => bang
                  puts bang
                  puts bang.backtrace
                  # puts @session.to_yaml
                end
              end
              res
            }
          end
        end
      end

      def updateCSRFToken(csrf_cache, request)
        # puts "=UPDATE CSRF TOKEN"
        # @session[:valid_csrf_tokens].to_yaml
        # puts request if request.site.nil?
        # puts "= = = = = = "
        @@csrf_lock.synchronize do
          #  if @session[:valid_csrf_tokens].has_key?(request.site)
          #    puts "* found token for site: #{request.site}"

          request.map!{ |line|
            res = line
            @session[:csrf_patterns].each do |pat|
              begin
                if line =~ /#{pat}/i then
                  key = Regexp.quote($1.upcase)
                  old_value = $2
                  # if @session[:valid_csrf_tokens][request.site].has_key?(key) then
                  if csrf_cache.has_key?(key) then
                    res = line.gsub!(/#{Regexp.quote(old_value)}/, csrf_cache[key])
                    if res.nil? then
                      res = line
                      puts "!!!could not update token (#{key})"
                    end
                    #     puts "->#{line}"

                    #end
                  end
                end
              rescue => bang
                puts bang
                puts bang.backtrace
                # puts @session.to_yaml
              end
            end
            res
          }
        end
        # end
      end

      # this function updates specific patterns of a request, e.g. CSRF Tokens
      # Parameters:
      # request - the request which has to be updated
      # cache - the value store of already collected key-value-pairs
      # patterns - pattern expressions, similar to session-id-patterns, e.g.  /name="(sessid)" value="([0-9a-zA-Z!-]*)"/
      def updateRequestPattern(request, cache, patterns)

        request.map!{ |line|
          res = line
          patterns.each do |pat|
            begin
              if line =~ /#{pat}/i then
                pattern_key = Regexp.quote($1.upcase)
                old_value = Regexp.quote($2)
                if cache.has_key?(sid_key) then
                  if not old_value =~ /#{cache[sid_key]}/ then # sid value has changed and needs update
                    #      print "S"
                    #    puts "+ update sid #{sid_key}"
                    #    puts "-OLD: #{old_value}"
                    #    puts "-NEW: #{@session[:valid_sids][request.site][sid_key]}"

                    #      puts "---"
                    # dummy = Regexp.quote(old_value)
                    res = line.gsub!(/#{old_value}/, cache[sid_key])
                    if not res then puts "!!!could not update sid (#{sid_key})"; end
                    #     puts "->#{line}"
                  end
                end
              end
            rescue => bang
              puts bang
              puts bang.backtrace
              # puts @session.to_yaml
            end
          end
          res
        }
      end

      def applySessionSettings(prefs)
        [ :update_valid_sids, :update_session, :update_contentlength, :valid_sids, :sid_patterns, :logout_signatures ].each do |v|
          @@settings[v] = prefs[v] if prefs[v]
        end
      end

    end
  end
end