# .
# quick_scan_dialog.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
require 'lib/gui/check_policy_menu'

module Watobo
  module Gui
    class QuickScanOptionsFrame < FXVerticalFrame
      def options()
        o = Hash.new
        o[:enable_logging] = @logScanChats.checked?
        o[:scan_log_dir] = @scanlog_dir_dt.value
        o[:csrf_tokens] = @csrf_patterns
        o[:csrf_request_ids] = @csrf_ids
        o[:update_csrf_tokens] = @csrfToken.checked?
        o[:use_orig_request] = @useOriginalRequest.checked?
          o
      end

      def initialize(owner, project=nil, prefs = {} )
        @project = project

        super(owner, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)

        @csrf_ids = []
        @csrf_patterns = []
          @scan_request = prefs[:target_request]
          
        # scan_opt_frame= FXVerticalFrame.new(self, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
        @useOriginalRequest = FXCheckButton.new(self, "Use Original Request", nil, 0, JUSTIFY_LEFT|JUSTIFY_TOP|ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP)
        @useOriginalRequest.checkState = true

        frame = FXGroupBox.new(self, "Logging", LAYOUT_SIDE_TOP|FRAME_GROOVE|LAYOUT_FILL_X, 0, 0, 0, 0)
        @logScanChats = FXCheckButton.new(frame, "Log Scan", nil, 0, JUSTIFY_LEFT|JUSTIFY_TOP|ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP)
        @logScanChats.checkState = false
        @logScanChats.connect(SEL_COMMAND) do |sender, sel, item|
          if @logScanChats.checked? then
            @scanlog_dir_text.enabled = true
            @scanlog_dir_label.enabled = true
            @scanlog_dir_btn.enable
          else
            @scanlog_dir_text.enabled = false
            @scanlog_dir_label.enabled = false
            @scanlog_dir_btn.disable
          end
        end

        @scanlog_dir_dt = FXDataTarget.new('')
        @scanlog_dir_dt.value = @project.scanLogDirectory() if File.exist?(@project.scanLogDirectory())
        @scanlog_dir_label = FXLabel.new(frame, "Scan-Log Directory:" )
        scanlog_frame = FXHorizontalFrame.new(frame,:opts => LAYOUT_FILL_X|LAYOUT_SIDE_TOP)
        @scanlog_dir_text = FXTextField.new(scanlog_frame, 20,
        :target => @scanlog_dir_dt, :selector => FXDataTarget::ID_VALUE,
        :opts => TEXTFIELD_NORMAL|LAYOUT_FILL_COLUMN|LAYOUT_FILL_X)
        @scanlog_dir_text.handle(self, FXSEL(SEL_UPDATE, 0), nil)
        @scanlog_dir_btn = FXButton.new(scanlog_frame, "Change")
        @scanlog_dir_btn.connect(SEL_COMMAND, method(:selectScanlogDirectory))

        @scanlog_dir_text.enabled = false
        @scanlog_dir_label.enabled = false
        @scanlog_dir_btn.disable

        frame = FXGroupBox.new(self, "One-Time-Token Settings", LAYOUT_SIDE_TOP|FRAME_GROOVE|LAYOUT_FILL_X, 0, 0, 0, 0)
        csrf_frame = FXHorizontalFrame.new(frame,:opts => LAYOUT_FILL_X|LAYOUT_SIDE_TOP, :padding => 0)
        @csrfToken = FXCheckButton.new(csrf_frame, "Update One-Time-Tokens", nil, 0, JUSTIFY_LEFT|JUSTIFY_TOP|ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP)
        @csrfToken.checkState = false
        @csrfToken.connect(SEL_COMMAND) do |sender, sel, item|
          if @csrfToken.checked? then
            @csrf_dialog_btn.enable
          else
            @csrf_dialog_btn.disable
          end
        end

        @csrf_dialog_btn = FXButton.new(csrf_frame, "O-T-T Settings")
        @csrf_dialog_btn.connect(SEL_COMMAND, method(:openCSRFTokenDialog))

        #  @csrf_dialog_text.enabled = false
        #  @csrf_dialog_label.enabled = false
        @csrf_dialog_btn.disable
      end

      private

      def openCSRFTokenDialog(sender, sel, item)
        csrf_dlg = CSRFTokenDialog.new(self, @project, @scan_request)
        if csrf_dlg.execute != 0 then
          @csrf_ids = csrf_dlg.getTokenScriptIds()
          @csrf_patterns = csrf_dlg.getTokenPatterns()

         # puts "* update CSRF Settings"
          puts @csrf_ids
          puts @csrf_patterns
          puts @scan_request
          
            @project.setCSRFRequest(@scan_request, @csrf_ids, @csrf_patterns)
          
          # puts csrf_ids.to_yaml
          # puts "= = ="
          # puts csrf_patterns.to_yaml

         # if @project
         #   @csrf_requests = []
         #   csrf_ids.each do |id|
         #     chat = @project.getChat(id)
         #     @csrf_requests.push chat.copyRequest
         #   end

          #end
          # save settings
          #  saveProjectSettings(@active_project)
          #  saveSessionSettings(@active_project)
        end
      end

      def selectScanlogDirectory(sender, sel, item)
        workspace_dt = FXFileDialog.getOpenDirectory(self, "Select Scanlog Directory", @scanlog_dir_dt.value)
        if workspace_dt != "" then
          if File.exists?(workspace_dt) then
            @scanlog_dir_dt.value = workspace_dt
            @scanlog_dir_text.handle(self, FXSEL(SEL_UPDATE, 0), nil)
          end
        end
      end

    end

    class QuickScanDialog < FXDialogBox

      include Responder

      attr :active_policy
      attr :selectedModules
      attr :options
      
      def initialize(parent, project=nil, prefs={} )
        super(parent, "Quick Scan", DECOR_ALL, :width => 300, :height => 400)
        # @active_policy = 'Default'
        @selectedModules = []
        @project = project
        FXMAPFUNC(SEL_COMMAND, ID_ACCEPT, :onAccept)

        base_frame = FXVerticalFrame.new(self, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        @switcher = FXSwitcher.new(base_frame,LAYOUT_FILL_X|LAYOUT_FILL_Y)

        @quickScanOptionsFrame = QuickScanOptionsFrame.new(@switcher, @project, prefs)
        @policyFrame = ChecksPolicyFrame.new(@switcher, project.active_checks, project.settings[:policy])

        # BUTTONS
        buttons_frame = FXHorizontalFrame.new(base_frame, :opts => LAYOUT_FILL_X)
        @startButton = FXButton.new(buttons_frame, "Start" ,
        :target => self, :selector => FXDialogBox::ID_ACCEPT,
        :opts => BUTTON_NORMAL|LAYOUT_RIGHT)
        @startButton.disable

        @nextButton = FXButton.new(buttons_frame, "Next" ,  nil, nil, :opts => BUTTON_NORMAL|LAYOUT_RIGHT)
        @nextButton.enable
        @nextButton.connect(SEL_COMMAND) do |sender, sel, item|
          if @switcher.current < @switcher.numChildren-1
            @switcher.current = @switcher.current+1

          end
          setButtons(@switcher.current)
        end

        @backButton = FXButton.new(buttons_frame, "Back" ,  nil, nil, :opts => BUTTON_NORMAL|LAYOUT_RIGHT)
        @backButton.disable
        @backButton.connect(SEL_COMMAND) do |sender, sel, item|
          if @switcher.current > 0
            @switcher.current = @switcher.current-1

          end
          setButtons(@switcher.current)
        end

        @cancelButton = FXButton.new(buttons_frame, "Cancel" ,
        :target => self, :selector => FXDialogBox::ID_CANCEL,
        :opts => BUTTON_NORMAL|LAYOUT_RIGHT)

        #@filterCombo.appendItem("ohne Filter", nil)

      end

      private

      def onAccept(sender, sel, event)
        @selectedPolicy = @policyFrame.policy_name
        @options = @quickScanOptionsFrame.options
        @selectedModules = @policyFrame.getSelectedModules()
        getApp().stopModal(self, 1)
        self.hide()
        return 1
      end

      def setButtons(index)
        case index
        when 0

          @nextButton.enable
          @backButton.disable
          @startButton.disable

        when 1
          # select session screen
          @nextButton.disable
          @backButton.enable
          @startButton.enable
        end
      end

    end
  end
end
