# .
# interceptor.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
require 'lib/httpparser'
require 'lib/core/session'
require 'lib/qShapers'
require 'lib/guiUtils'
require 'lib/core/project'
require 'lib/http_socket'
require 'lib/core/watobo_constants'
require 'socket'
require 'fox16'
require 'pathname'

include Fox

module Watobo
  module Core

    INTERCEPT_NONE = 0x00
    INTERCEPT_REQUEST = 0x01
    INTERCEPT_RESPONSE = 0x02
    INTERCEPT_BOTH = 0x03

    INTERCEPT_DEFAULT_PORT = 8081
    #   I N T E R C E P T P R O X Y
    #
    class InterceptProxy
      attr_accessor :mode

      attr_accessor :contentLength
      attr_accessor :contentTypes
      attr_accessor :target
      attr_accessor :www_auth
      attr :port
      def subscribe(event, &callback)
        (@event_dispatcher_listeners[event] ||= []) << callback
      end

      def clearEvents(event)
        @event_dispatcher_listener[event].clear
      end

      def notify(event, *args)
        if @event_dispatcher_listeners[event]
          @event_dispatcher_listeners[event].each do |m|
            m.call(*args) if m.respond_to? :call
          end
        end
      end

      def addPreview(response)
        preview_id = Digest::MD5.hexdigest(response.join)
        @preview[preview_id] = response
        return preview_id
      end

      def matchContentType?(content_type)
        @contentTypes.each do |p|
          return true if content_type =~ /#{p}/
        end
        return false
      end

      def isPassThrough?(request, response, s_sock, c_sock)
        begin
          reason = nil
          clen = response.content_length

          if matchContentType?(response.content_type) then
            # first forward headers
            c_sock.write response.join
            reason = []
            reason.push "---> WATOBO: PASS_THROUGH <---"
            reason.push "Reason: Content-Type = #{response.content_type}"
          elsif clen > @contentLength
            # puts "PASS-THROUGH: #{response.content_length}"
            c_sock.write response.join
            reason = []
            reason.push "---> WATOBO: PASS_THROUGH <---"
            reason.push "Reason: Content-Length > #{@contentLength} (#{response.content_length})"
          end

          return false if reason.nil?
          reason.push "* DO MANUAL REQUEST TO GET FULL RESPONSE *"
          response.push reason.join("\n")
          chat = Watobo::Core::Chat.new(request, response, :source => CHAT_SOURCE_INTERCEPT)
          notify(:new_interception, chat)

          pass_through(s_sock, c_sock, clen)
          #  puts "* Close Server Socket..."
          closeSocket(c_sock)
          #  puts "* Close Client Socket..."
          closeSocket(s_sock)
          #  puts "... done."
          true
        rescue => bang
          puts bang
          puts bang.backtrace
          return false
        end
      end

      def stop()
        @intercept_srv.close
      end

      def closeSocket(socket)
        #puts socket.class
        begin
          #if socket.class.to_s =~ /SSLSocket/
          if socket.is_a? OpenSSL::SSL::SSLSocket
            socket.io.shutdown(2)
          else
            socket.shutdown(2)
          end
          socket.close
        rescue => bang
         # puts bang
         # puts bang.backtrace
        end
      end

      def pass_through(server, client, maxbytes = 0)
        buf = ''
        print "[~"
        bytes_read = 0
        while buf
          begin
            #timeout(2) do
            buf = nil
            buf = server.readpartial(2048)
            #end
          rescue EOFError
            #client.write buf if buf
            #print "~]"
            return if buf.nil?
          rescue Errno::ECONNRESET
            puts "!!! ERROR (Reset): reading body"
            puts "* last data seen on socket: #{buf}"
            return if buf.nil?
          rescue Timeout::Error
            puts "!!! ERROR (Timeout): reading body"
            puts "* last data seen on socket:"
            #client.write buf if buf
            print "!~]"
            return if buf.nil?
          rescue => bang
            puts "!!! Error (???) reading body:"
            puts bang
            puts bang.class
            puts bang.backtrace.join("\n")
            puts "* last data seen on socket:"
            print "~]"
            #client.write buf if buf
            return
          end

          begin
            return if buf.nil?
            print "~"
            client.write buf
            bytes_read += buf.length
            # puts "##{bytes_read} of #{maxbytes}"
            if maxbytes > 0 and bytes_read >= maxbytes
              print "~]"
              return
            end

          rescue Errno::ECONNRESET
            print "~x]"
            #  puts "!!! ERROR (Reset): reading body"
            #  puts "* last data seen on socket: #{buf}"
            return
          rescue Errno::ECONNABORTED
            print "~x]"
            return
          rescue Errno::EPIPE
            print "~x]"
            return
          rescue => bang
            puts "!!! Error (???) in Client Communication:"
            puts bang
            puts bang.class
            puts bang.backtrace.join("\n")
            return
          end
        end
      end

      def read_request(socket)
        request = []
        # read http header lines
        session = socket
        # puts "* read header ..."
        Watobo::Framework::HTTP.read_header(socket) do |line|
          request.push line
        end

        if request.first =~ /^CONNECT (.*):(\d{1,5}) HTTP\/1\./ then
          target = $1
          tport = $2
          # puts request.first
          #  print "\n* CONNECT: #{method} #{target} on port #{tport}\n"

          socket.print "HTTP/1.0 200 Connection established\r\n" +
          "Proxy-connection: Keep-alive\r\n" +
          "Proxy-agent: WATOBO-Proxy/1.1\r\n" +
          "\r\n"
          bscount = 0 # bad handshake counter
          #  puts "* wait for ssl handshake ..."
          begin
            ssl_socket = OpenSSL::SSL::SSLSocket.new(socket, @ctx)
            ssl_socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
            #ssl_socket.sync_close = true
            ssl_socket.sync = true
            # puts ssl_socket.methods.sort

            ssl_session = ssl_socket.accept
          rescue => bang

            #puts ssl_session
            #if not ssl_session then bscount += 1;end
            #if bscount > 10 then
            #  puts "!!! Error: SSL-Handshake with Client/Browser"
            #  puts bang
            return nil, socket
            #end
            #retry
          end
          session = ssl_session
          # puts "* ssl ok!"
          # now read ssl request header
          request = []
          Watobo::Framework::HTTP.read_header(session) do |line|
            request.push line
          end

          return nil, session if not request.first

          request.first.gsub!(/(^[^[:space:]]{1,})( )(\/.*)/, "\\1 https://#{target}:#{tport}\\3")
        end
        #puts request
        request.extend Watobo::Parser::Url
        request.extend Watobo::Parser::Web10
        request.extend Watobo::Framework::Shaper::Web10

        clen = request.content_length
        if  clen > 0 then
          body = ""
          Watobo::Framework::HTTP.read_body(session) do |data|
            body += data
            break if body.length == clen
          end
          request.push body
        end

        return request, session
      end

      #
      # R U N
      #

      def run(settings={})
        begin
          @port = settings[:port] if settings[:port]
          @contentLength = settings[:content_length] if settings[:content_length]
          @contentTypes = settings[:content_types] if settings[:content_types]

          @intercept_srv = TCPServer.new("127.0.0.1", @port)
          @intercept_srv.setsockopt(Socket::SOL_SOCKET, Socket::SO_KEEPALIVE, 1)
        rescue => bang
          puts "\n!!!Could not start InterceptProxy"
          puts bang
          return nil
        end
        puts "\n* Interception started on port #{@port}"
        session_list = []

        Thread.new(@intercept_srv) { |server|
          while (new_session = server.accept)
            #  new_session.sync = true
            Thread.new(new_session) { |session|
              begin
                request, c_sock = read_request(session)

                if not request then
                  c_sock.close
                  next
                  #return
                end
                puts "*[I] #{request.url}"
                #puts "-"
              rescue => bang
                puts "!!! Error reading client request "
                puts bang
                next
                #break
              end

              # check if preview is requested
              if request.first =~ /WATOBOPreview=([0-9a-zA-Z]*)/ then
                puts "* preview requested ..."
                puts request.url

                hashid = $1
                response = @preview[hashid]

                if response then
                  c_sock.write response.join
                  closeSocket(c_sock)
                end
                next
              end

              # no preview, check if interception request is turned on
              if @target and @mode & INTERCEPT_REQUEST > 0 then
                @awaiting_requests += 1

                if @target.respond_to? :modifyRequest
                  #  puts "*INTERCEPT REQUEST"
                  #  puts @target
                  #notify(:modify_request, request, Thread.current)
                  #  puts "send request to target"
                  @target.modifyRequest(request, Thread.current)
                  Thread.stop
                else
                  p "! no target for editing request"
                end
                @awaiting_requests -= 1
              end
              # req, resp = @sender.sendRequest(request, :update_sids => false, :update_session => false, :update_contentlength => true)

              #p "getHTTPHeader"
              #s_sock, req, resp = @sender.getHTTPHeader(request, :update_sids => true, :update_session => false, :update_contentlength => true)
              s_sock, req, resp = @sender.sendHTTPRequest(request, :update_sids => true, :update_session => false, :update_contentlength => true, :www_auth => @www_auth)

              if s_sock.nil? then
                #TODO: Do some error handling here
                #puts "! Could not connect"
                c_sock.print"HTTP\/1.1 504 Gateway\r\n\r\n#{resp}"
                closeSocket(c_sock)
                next
              end

              # check if response is ready should be passed throug
              next if isPassThrough?(req, resp, s_sock, c_sock)
              #p "no pass-through"

              begin
                puts "* got response status: #{resp.status}"
                rs = resp.status
                if rs =~ /^(401|407)/ then
                  resp.extend Watobo::Framework::Shaper::Web10
                  
                  auth_type = AUTH_TYPE_NONE
                  resp.each do |rl|
                    auth_type = AUTH_TYPE_BASIC if rl =~ /^WWW-Authenticate: Basic/
                    auth_type = AUTH_TYPE_NTLM if rl =~ /^WWW-Authenticate: NTLM/
                  end
                  # when auth type not basic assume it's ntlm -> ntlm credentials must be set in watobo
                  unless auth_type == AUTH_TYPE_NONE
                    if rs =~ /^401/ then
                      resp.push "WATOBO: Server requires (NTLM) authorization, please set WWW_Auth Credentials!"
                      resp.shift
                      resp.unshift "HTTP/1.1 200 OK\r\n"
                    else
                      resp.push "WATOBO: Proxy requires (NTLM) authorization, please set Proxy Credentials!"
                      resp.shift
                      resp.unshift "HTTP/1.1 200 OK\r\n"
                    end
                  else
                    resp.push "WATOBO: Unknown authorization type.<br><br>\r\n" + resp.join("<br>\r\n")
                                          resp.shift
                                          resp.unshift "HTTP/1.1 200 OK\r\n"
                    resp.fix_content_length
                   
                  end
                else
                  @sender.readHTTPBody(s_sock, resp, req, :update_sids => true)
                end
              rescue => bang
                puts bang
                puts bang.backtrace
                puts "* Error sending request"
              end

              begin
                if @target and @mode & INTERCEPT_RESPONSE > 0 then
                  if resp.content_type =~ /text/ or resp.content_type =~ /application\/javascript/ then
                    if @target.respond_to? :modifyResponse
                      @target.modifyResponse(resp, Thread.current)
                      Thread.stop
                    else
                      p "! no target for editing response"
                    end
                  end
                end

                c_sock.write resp.join
                closeSocket(c_sock)
              rescue Errno::ECONNRESET
                print "x"
                #  puts "!!! ERROR (Reset): reading body"
                #  puts "* last data seen on socket: #{buf}"
                #return
              rescue Errno::ECONNABORTED
                print "x"
                #return
              rescue => bang
                puts "!!! Error (???) in Client Communication:"
                puts bang
                puts bang.class
                puts bang.backtrace.join("\n")
                #return
              end
              #if req then # when req == nil an error occured -> no chat necessary
              chat = Chat.new(request, resp, :source => CHAT_SOURCE_INTERCEPT)
              notify(:new_interception, chat)
              # end
              #else
              #  puts "* got no response"
              #  session_new.close
            }
          end
        }
      end

      def initialize(settings={})
        @event_dispatcher_listeners = Hash.new
        begin
          puts
          puts "=== Initialize Interceptor/Proxy ==="
          #   @project = project
          #   @settings = settings
          # @port = @settings[:intercept_port]
          puts settings.to_yaml

          @mode = INTERCEPT_NONE
          @www_auth = {}
          @target = nil
          @sender = Watobo::Core::Session.new(@target)

          @port = INTERCEPT_DEFAULT_PORT

          @contentLength = 100000
          @contentTypes = [ 'application\/audio',
            'application\/video',
            'application\/image',
            'application\/pdf',
            'application\/octet',
            'application\/.*flash',
            'image\/'
          ]

          @port = settings[:port] if settings[:port]
          if pt_settings = settings[:pass_through] then
            @contentLength = pt_settings[:content_length] if pt_settings[:content_length]
            @contentTypes = pt_settings[:content_types] if pt_settings[:content_types]
          end

          @awaiting_requests = 0
          @awaiting_responses = 0

          @preview = Hash.new
          @preview['ProxyTest'] = ["HTTP/1.0 200 OK\r\nServer: Watobo-Interceptor\r\nConnection: close\r\nContent-Type: text/html; charset=iso-8859-1\r\n\r\n<html><body>PROXY_OK</body></html>"]

          # p @settings[:certificate_path]
          # p @settings[:cert_file]
          # p @settings[:key_file]
          cert_file = File.join($watobo_base, settings[:certificate_path], settings[:cert_file])
          key_file = File.join($watobo_base, settings[:certificate_path], settings[:key_file])

          #  @ctx = OpenSSL::SSL::SSLContext.new('SSLv23_server')
          @ctx = OpenSSL::SSL::SSLContext.new()

          @ctx.cert = OpenSSL::X509::Certificate.new(File.read(cert_file))
          #  @ctx.key = OpenSSL::PKey::DSA.new(File.read(key_file))
          @ctx.key = OpenSSL::PKey::RSA.new(File.read(key_file))
          @ctx.verify_mode = OpenSSL::SSL::VERIFY_NONE
          @ctx.timeout = 10
          #  @ctx.ciphers = nil # ['TLSv1/SSLv3', 56, 56 ]

        rescue => bang
          puts "!!!could not read certificate files:"
          puts bang
          puts bang.backtrace
        end

      end

      private

    end

    #
    #--------------------------------------------------------------------------------------------
    #

    # class FXInterceptor < FXDialogBox
    class FXInterceptor < FXTopWindow

      include Responder
      include Watobo::Core
      def execute
        create
        show(PLACEMENT_SCREEN)
        # getApp().runModalFor(self)
      end

      def compileNewResponse(response)
        resp = response.parseRequest()
        resp.extend Watobo::Parser::Web10
        resp.extend Watobo::Framework::Shaper::Web10
        resp.fixupContentLength()

        return resp
      end

      def onInterceptChanged(sender, sel, ptr)
        begin
          if not @interceptor.nil? then
            @interceptor.mode = @intercept_response.checked? ? INTERCEPT_RESPONSE : 0
            @interceptor.mode |= @intercept_request.checked? ? INTERCEPT_REQUEST : 0
          end
        rescue => bang
          puts bang
          puts bang.backtrace
        end
      end

      def onAcceptChanges(sender, sel, ptr)
        if @tabBook.current == 0 then

          begin
            request = @request_list.first
            if not request.nil?
              request[:request].clear
              request[:request].concat @requestbox.parseRequest
              request[:thread].run
              @request_list.shift
              getNextRequest()
            end
          rescue => bang
            puts "!!! Error"
            puts bang
          end
        else
          begin
            response = @response_list.first
            if not response.nil?
              response[:response].clear
              new_response = compileNewResponse(@responsebox)
              response[:response].concat new_response
              response[:thread].run

              @response_list.shift
              getNextResponse()
            end
          rescue => bang
            puts "!!! Error"
            puts bang
          end
        end

      end

      def onDrop(sender, sel, ptr)
        if @tabBook.current == 0 then
          request = @request_list.first
          if request
            request[:request].clear
            request[:thread].run
            @request_list.shift
          end
          getNextRequest()
        else
          response = @response_list.first
          if response
            response[:response].clear
            response[:thread].run
            @response_list.shift
          end
          getNextResponse()

        end
      end

      def getNextResponse
        @edit_response = @response_list.first
        if not @edit_response.nil? then
          #   puts "* edit next response"
          @responsebox.setText(@edit_response[:response])
          @response_tab.text = "Response (#{@response_list.length})"
        else
          disable_buttons()
          @responsebox.setText('')
          # @response_tab.text = "Response (0)"
        end
        @response_tab.text = "Response (#{@response_list.length})"
      end

      def getNextRequest
        request = @request_list.first
        if not request.nil? then
          # puts "* edit next request"
          @requestbox.setText(request[:request])
          @request_tab.text = "Request (#{@request_list.length})"
        else
          disable_buttons()
          @requestbox.setText('')
          # @request_tab.text = "Request (0)"
        end
        @request_tab.text = "Request (#{@request_list.length})"
      end

      def onDiscard(sender, sel, ptr)
        if @tabBook.current == 0 then
          request = @request_list.first
          request[:thread].run if request
          @request_list.shift
          getNextRequest()
        else
          response = @response_list.first
          response[:thread].run if response
          @response_list.shift
          getNextResponse()
        end
      end

      def onDiscardAll(sender, sel, ptr)

      end

      def onHide
        #  puts "* hiding interceptor"
        @interceptor.mode = INTERCEPT_NONE
        @mutex.synchronize {
          @cv.signal
        }
      end

      def releaseAll()
        #  puts "* closing interceptor"
        #  puts "* disable all interceptions"
        #@project.intercept_request = false
        #@project.intercept_response = false
        #  puts "* release all interceptions ..."
        @request_list.each do |ir|
          ir[:thread].run
        end
        @response_list.each do |ir|
          ir[:thread].run
        end

      end

      def onClose(sender, sel, ptr)
        releaseAll()
        #getApp().stopModal(self, 1)
        self.hide()
      end

      def modifyRequest(request, thread)
        #  puts "* FXInterceptor: modifyRequest"
        new_request = {
          :request => request,
          :thread => thread
        }
        enable_buttons()
        @request_list.push new_request
        @request_tab.text = "Request (#{@request_list.length})"

        if @request_list.length == 1 then
          @edit_request = new_request
          @requestbox.setText(request)
        end

        enable_buttons()
      end

      def modifyResponse(response, thread)
        response.extend Watobo::Parser::Web10
        response.extend Watobo::Framework::Shaper::Web10
        response.fixupContentLength()

        new_response = {
          :response => response,
          :thread => thread
        }
        enable_buttons()
        @response_list.push new_response
        @response_tab.text = "Response (#{@response_list.length})"
        if @response_list.length == 1 then
          #@edit_response = new_response
          @responsebox.setText(response)
          #  @tabBook.setCurrent(1)
          enable_buttons()

        end

      end

      def enable_buttons
        @accept_button.enabled = true
        @discard_button.enabled = true
        @drop_button.enabled = true
      end

      def disable_buttons
        @accept_button.enabled = false
        @discard_button.enabled = false
        @drop_button.enabled = false
      end

      def initialize(owner, interceptor, opts)
        # Invoke base class initialize function first

        super( owner, 'Interceptor', nil, nil, DECOR_ALL|DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE, 0, 0, 500, 500, 0, 0, 0, 0, 0, 0)
        self.connect(SEL_CLOSE, method(:onClose))

        @interceptor = interceptor

        @request_list = []
        @response_list = []

        @edit_request = nil
        @edit_response = nil

        # initial frame setup
        mr_splitter = FXSplitter.new(self, LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_VERTICAL|SPLITTER_REVERSED|SPLITTER_TRACKING)

        top_frame = FXVerticalFrame.new(mr_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y||LAYOUT_FIX_HEIGHT|LAYOUT_BOTTOM,:height => 500)
        top_splitter = FXSplitter.new(top_frame, LAYOUT_FILL_X|SPLITTER_HORIZONTAL|LAYOUT_FILL_Y|SPLITTER_TRACKING)

        log_frame = FXVerticalFrame.new(mr_splitter, :opts => LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM,:height => 100)

        filter_frame = FXVerticalFrame.new(top_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y||LAYOUT_FIX_HEIGHT|LAYOUT_BOTTOM)
        view_frame = FXVerticalFrame.new(top_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y||LAYOUT_FIX_HEIGHT|LAYOUT_BOTTOM)

        @tabBook = FXTabBook.new(view_frame, nil, 0, LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_RIGHT)
        button_frame = FXHorizontalFrame.new(view_frame, LAYOUT_FILL_X|LAYOUT_FILL_Y)
        @tabBook.connect(SEL_COMMAND) { |sender, sel, item|
          case item
          when 0
            enable_buttons if @request_list.length > 0
          when 1
            enable_buttons if @response_list.length > 0
          end

        }
        @request_tab = FXTabItem.new(@tabBook, "Request", nil)
        request_frame_outer = FXVerticalFrame.new(@tabBook, LAYOUT_FILL_X|LAYOUT_FILL_Y)
        request_frame = FXVerticalFrame.new(request_frame_outer, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK, :padding=>0)
        @requestbox = Watobo::Gui::RequestEditor.new(request_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
        @requestbox.editable = true

        @response_tab = FXTabItem.new(@tabBook, "Response", nil)
        response_frame_outer = FXVerticalFrame.new(@tabBook, LAYOUT_FILL_X|LAYOUT_FILL_Y)
        response_frame = FXVerticalFrame.new(response_frame_outer, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK, :padding=>0)
        @responsebox = Watobo::Gui::RequestEditor.new(response_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y )
        @responsebox.editable = true

        FXLabel.new(filter_frame, "Intercept:" )
        @intercept_request = FXCheckButton.new(filter_frame, "Requests", nil, 0,
        ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP)
        @intercept_request.connect(SEL_COMMAND, method(:onInterceptChanged))

        @intercept_response = FXCheckButton.new(filter_frame, "Response", nil, 0,
        ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP)
        @intercept_response.connect(SEL_COMMAND, method(:onInterceptChanged))

        case @interceptor.mode
        when INTERCEPT_REQUEST
          @intercept_request.checkState = true
          @intercept_response.checkState = false
        when INTERCEPT_RESPONSE
          @intercept_response.checkState = true
          @intercept_request.checkState = false
        when INTERCEPT_BOTH
          @intercept_request.checkState = true
          @intercept_response.checkState = true
        when INTERCEPT_NONE
          @intercept_request.checkState = false
          @intercept_response.checkState = false
        end

        @accept_button = FXButton.new(button_frame, "Accept", nil, nil, 0, FRAME_RAISED|FRAME_THICK|LAYOUT_LEFT)
        @accept_button.connect(SEL_COMMAND, method(:onAcceptChanges))

        @discard_button = FXButton.new(button_frame, "Discard", nil, nil, 0, FRAME_RAISED|FRAME_THICK|LAYOUT_LEFT)
        @discard_button.connect(SEL_COMMAND, method(:onDiscard))

        @drop_button = FXButton.new(button_frame, "Drop", nil, nil, 0, FRAME_RAISED|FRAME_THICK|LAYOUT_LEFT)
        @drop_button.connect(SEL_COMMAND, method(:onDrop))

        log_text_frame = FXVerticalFrame.new(log_frame, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK, :padding=>0)
        @log_viewer = FXText.new(log_text_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)

        disable_buttons()

      end
    end
  end
end