# .
# qGui.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
require 'fox16'
require 'lib/core/watobo_constants'
require 'lib/utils_include.rb'
require 'lib/qLoginWizzard.rb'
require 'lib/qManualRequest.rb'
require 'lib/encoders.rb'
require 'lib/guiUtils.rb'
require 'lib/interceptor.rb'
require 'lib/transcoder.rb'
require 'lib/chatviewer.rb'
require 'lib/requesteditor.rb'
require 'lib/gui/project_wizzard.rb'
require 'lib/gui/workspace_dialog.rb'
require 'lib/gui/findings_tree.rb'
require 'lib/gui/sites_tree.rb'
require 'lib/gui/progress_window'
require 'lib/gui/proxy_dialog.rb'
require 'lib/gui/fuzzer_gui'
require 'lib/gui/check_policy_menu'
require 'lib/gui/dashboard'
require 'lib/gui/plugin_board'
require 'lib/gui/certificate_dialog'
require 'lib/gui/finding_info'
require 'lib/gui/start_full_scan_dialog'
require 'lib/gui/quick_scan_dialog'
require 'lib/gui/session_management_dialog'
require 'lib/gui/preferences_dialog'
require 'lib/gui/browser_preview'
require 'lib/gui/status_bar'
require 'lib/gui/about_watobo'
require 'lib/gui/edit_scope_dialog'
require 'lib/gui/scanner_settings_dialog'
require 'lib/gui/interceptor_settings_dialog'
require 'lib/gui/confirm_scan_dialog'
require 'lib/gui/www_auth_dialog'
require 'lib/gui/edit_comment'
require 'lib/core/active_check.rb'

include Fox
include Watobo::Core

module Watobo
  module Gui
    def getList(pattern)
      vals = pattern.split(",")

      result = []
      vals.each do |v|
        v.strip!
        if v =~ /^(\d+)$/ then
          result.push $1.to_i
        elsif v =~ /^(\d+)-(\d+)$/
          start = $1
          stop = $2
          dummy = (start..stop).to_a
          result.concat dummy
        end
      end
      result.uniq!
      return result
    end

    #===================================================================================================================

    ####################################################################################################################
    # M A I N   A P P L I C A T I O N   W I N D O W
    #
    class WatoboMainWindow < FXMainWindow
      include Watobo::Gui
      include Watobo::Gui::Utils

      attr :interceptor
      attr :default_settings
      attr :watobo_base
      attr :active_project
      attr :iproxy
      # Handles SEL_CLOSE from the main window
      def open_manual_request_toolkit(chat)
        begin
          mrtk = QXManualRequest.new(FXApp.instance, @active_project, chat)

          mrtk.create

          mrtk.subscribe(:show_browser_preview) { |request, response|
            openBrowser(request, response)
          }

          mrtk.subscribe(:new_chat) { |chat|
            # p "<- :new_chat"
            @active_project.addChat(chat)
          }
          mrtk.show(Fox::PLACEMENT_SCREEN)
        rescue => bang
          puts "!!! could not open manual request"
          puts bang
        end
      end

      private

      def chatIsFiltered?(chat)
        begin
          #if @active_project and @active_project.settings[:site_filter]
          #  #puts chat.request.url
          #puts @active_project.settings[:site_filter]
          #  return false if @active_project.settings[:site_filter] != '' and chat.request.url =~ /^http(s)?:\/\/#{Regexp.quote(@active_project.settings[:site_filter])}/
          #  return true
          # end
          if @table_option_scope.checked?
            return true if !@active_project.siteInScope?(chat.request.site)
          end

          if @table_option_hidetested_cb.checked?
            return true if chat.tested?
          end

          unless @doctype_TableFilter.include?(chat.request.doctype)
            if @table_filter == "" then
              return false
            elsif @foption_req.checked? and chat.request.join =~ /#{@table_filter}/i then
              return false
            elsif @foption_res.checked? and chat.response.join =~ /#{@table_filter}/i then
              return false

            elsif chat.request.first =~ /#{@table_filter}/i then
              return false
            end
          end
          return true
        rescue => bang
          puts "!!!ERROR: could not add chat to table:"
          puts chat.id
          puts bang
          puts bang.backtrace
          return false
        end
      end

      #def loadDefaultS
      def saveDefaultSettings(update_settings={})
        settings= Hash.new
        [:history, :workspace_path, :interceptor_prefs, :forwarding_proxy, :project_settings_file_ext, :session_settings_file_ext].each do |k|
          settings[k] = @settings[k]
          settings[k] = update_settings[k] if update_settings.has_key?(k)
        end
        # clean up few settings
        settings[:forwarding_proxy][:name] = ''
        # puts "* saveDefaultSettings #{@default_settings_file}"
        # puts settings.to_yaml

        Watobo::Utils.save_settings(@default_settings_file, settings)
      end

      def saveSessionSettings(project=nil)
        if project then
          settings = Hash.new
          [:logout_signatures, :non_unique_parms, :login_chat_ids, :excluded_chats, :project_name, :session_name, :csrf_request_ids ].each do |k|
            settings[k] = Hash.new
            settings[k] = project.settings[k] if project.settings.has_key?(k)
          end
          file = @projects[project][:session_file]
          Watobo::Utils.save_settings(file, settings)
        end
      end

      def saveProjectSettings(project=nil)
        if project then
          ps = YAML.load(YAML.dump(project.settings))
          settings = Hash.new
          [:forwarding_proxy, :policy, :scope, :www_auth ].each do |k|
            settings[k] = Hash.new
            settings[k].update ps[k] if ps.has_key?(k)
          end

          fps = settings[:forwarding_proxy][:list]
          fps.each_key do |p|
            #TODO: implement encryption with master-password
            next if fps[p].is_a? String
            next unless fps[p].has_key?(:credentials)
            fps[p][:credentials][:password] = ''
          end

          was = settings[:www_auth]
          was.each_key do |p|
            #TODO: implement encryption with master-password
            next if was[p].is_a? String
            was[p][:password] = ''
          end

          settings[:project_name] = ps[:project_name]
          settings[:custom_error_patterns] = ps[:custom_error_patterns]
            settings[:csrf_patterns] = ps[:csrf_patterns] unless ps[:csrf_patterns].nil?

          file = @projects[project][:project_file]
          Watobo::Utils.save_settings(file, settings)
        end
      end

      def updateRequestTable(project)
        uniqueHashes = Hash.new
        # first initialize the table
        @chatTable.clearConversation()
        chats = project.chats
        # chats = project.chatsInScope(nil, nil) if @table_option_scope.checked?
        chats.each do |chat|
          if chatIsFiltered?(chat) == false
            uh = @active_project.uniqueRequestHash(chat.request)
            if @table_option_unique.checked? and not uniqueHashes.has_key?(uh)
              @chatTable.addChat(chat)
              uniqueHashes[uh] = chat
            elsif not @table_option_unique.checked?
              @chatTable.addChat(chat)
            else
              # puts "* skipping chat[#{chat.id}] - hash(#{uh}) already exists"
            end

          end
        end
      end

      #
      # SHOW CHAT
      #
      def showChat(chat)
        @mre_button.enabled = true
        @fuzz_button.enabled = true
        @bv_button.enabled = true

        @request_viewer.setText(chat.request)
        @last_request = chat.request
        @response_viewer.setText(chat.response)
        @last_response = chat.response
        @switcher.current=0
        @lastViewed = chat
        src = case chat.source
        when CHAT_SOURCE_INTERCEPT
          "Interceptor"
        when CHAT_SOURCE_PROXY
          "Proxy"
        when CHAT_SOURCE_MANUAL
          "Manual"
        when CHAT_SOURCE_FUZZER
          "Fuzzer"
        end
        @quickViewTitle.text = "Chat-ID: #{chat.id} (#{src})"
        @quickViewSubTitle.text = ""
      end

      #
      # SHOW VULN
      #
      def showVulnerability(vuln)
        @mre_button.enabled = true
        @fuzz_button.enabled = true
        @bv_button.enabled = true

        @chatTable.killSelection()
        @request_viewer.setText(vuln.request)
        @last_request = vuln.request
        @response_viewer.setText(vuln.response)
        @last_response = vuln.response

        @lastViewed = vuln
        if vuln.details[:check_pattern] then

          pattern = vuln.details[:check_pattern].strip

          @request_viewer.highlight(pattern)
        end

        if vuln.details[:proof_pattern] then
          pattern = vuln.details[:proof_pattern].strip
          @response_viewer.highlight(pattern)
        end
        @switcher.current = 0

        @quickViewTitle.text = "Finding: #{vuln.details[:class]}"
        @quickViewSubTitle.text = "Module: #{vuln.details[:module].gsub(/watobo::modules::/i,'')}"

      end

      def addFinding(finding)
        @findings_tree.addFinding(finding)
      end

      def openSessionManagement(sender, sel, item)
        smdlg = SessionManagementDialog.new(self, @active_project)
        if smdlg.execute != 0 then
          ids = smdlg.getLoginScriptIds()
          sidpatterns = smdlg.getSidPatterns()
          logout_signatures = smdlg.getLogoutSignatures()
          if @active_project
            @active_project.setLoginChatIds(ids)
            @active_project.setSidPatterns(sidpatterns)
            @active_project.setLogoutSignatures(logout_signatures)
          end
          # save settings
          saveProjectSettings(@active_project)
          saveSessionSettings(@active_project)
        end
      end

      def openPreferencesDialog(sender, sel, ptr)
=begin
        prefdlg = PreferencesDialog.new(self, @settings)
        if prefdlg.execute != 0
          ip = @settings[:interceptor_prefs]
          if ip[:intercept_port] != prefdlg.settings[:intercept_port] then
            FXMessageBox.information(self,MBOX_OK,"Intercept Port Changed", "Please restart WATOBO!")
          end
          @settings.update(prefdlg.settings)
          puts prefdlg.settings

          saveDefaultSettings()
          if @active_project then
            @active_project.settings.update(@settings)
            @active_project.saveProjectSettings()
          end

        end
=end
      end

      def openCADialog(sender, sel, item)
        cadlg = CertificateDialog.new(self, @active_project)
        if cadlg.execute != 0

        end
      end

      def openWwwAuthDialog()
        if @active_project.nil?
          FXMessageBox.information(self,MBOX_OK,"No Project Defined", "Create Project First")
        else
          auth_settings = {}
          w3adlg = Watobo::Gui::WwwAuthDialog.new(self, @active_project)
          if w3adlg.execute != 0
            #puts "* New WWW-Authentication"
            #puts @active_project.getWwwAuthentication().to_yaml
            saveProjectSettings(@active_project)
            @iproxy.www_auth = @active_project.settings[:www_auth]
          end
        end

      end

      def openFuzzer(chat)
        begin
          fuzzer = FuzzerGui.new(FXApp.instance, @active_project, chat)
          fuzzer.create
          fuzzer.show(Fox::PLACEMENT_SCREEN)
        rescue => bang
          puts "!!! could not open fuzzer"
          puts bang
        end
      end

      def addChat(chat)
        # addChatToTable(chat) if chatIsFiltered?(chat) == false
        @chatTable.addChat(chat) if chatIsFiltered?(chat) == false
        @sites_tree.addChat(chat)
      end

      def showPassiveCheckStatus
        @switcher.current=2
        @dashboard.tabBook.current = 1
      end

      def showActiveCheckStatus
        @switcher.current=2
        @dashboard.tabBook.current = 2
      end

      def showDashboard(sender, sel, ptr)
        #@switcher.current=2
        begin
          @switcher.setCurrent(2, true)
        rescue
          puts "no dashboard available yet!"
        end
      end

      def showLogs(sender, sel, ptr)
        @switcher.current=1
      end

      def showConversation(sender=nil, sel=nil, item=nil)
        @switcher.current=0
      end

      def openBrowser(request, response)
        begin
          @browserView.show(request, response)
        rescue => bang
          puts "!!! PROPBLEM !!"
          puts bang
          case bang
          when /JSSH_CONNECT_ERROR/i
            FXMessageBox.information(self, MBOX_OK, "JSSH Missing", "It seem that the Firefox JSSH extension is not installed,\nwhich is required in order to use the BrowserPreview.\nPlease read the installation instruction in the README\n or online at http://watobo.sourceforge.net.")
          else
            FXMessageBox.information(self, MBOX_OK, "Proxy Settings", "Your Browser does not use WATOBO (127.0.0.1:#{@iproxy.port}) as its proxy.\nSo you can't use the Browser-View feature.\nPlease change your proxy settings and try it again!")
          end
        end
      end

      def updateTreeLists()
        @findings_tree.reload()
        @sites_tree.reload()
      end

      def openTranscoder(text2transcode)
        transcoder = FXTranscoder.new(FXApp.instance, text2transcode)
        transcoder.create
        transcoder.show(Fox::PLACEMENT_SCREEN)
      end

      def showFindingDetails(finding)
        #p "* show finding details"
        @finding_info.showInfo(finding)
        @switcher.setCurrent(3, true)
      end

      def onShowPlugins(sender, sel, item)
        begin
          @switcher.setCurrent(4, true)
          @pluginboard.updateBoard(@active_project)
        rescue
          puts "no dashboard available yet!"
        end
      end

      def useSmallIcons()
        @findings_tree.useSmallIcons()
        @sites_tree.useSmallIcons()
        @chatTable.setNewFont( "helvetica", GUI_SMALL_FONT_SIZE)
        @request_viewer.setFontSize(GUI_SMALL_FONT_SIZE)
        @response_viewer.setFontSize(GUI_SMALL_FONT_SIZE)
      end

      def useRegularIcons()
       
        @findings_tree.useRegularIcons()
        @sites_tree.useRegularIcons()
        @chatTable.setNewFont("helvetica", GUI_REGULAR_FONT_SIZE)
        @request_viewer.setFontSize(GUI_REGULAR_FONT_SIZE)
        @response_viewer.setFontSize(GUI_REGULAR_FONT_SIZE)
      end

      def onOpenTranscoder(sender, sel, item)

        openTranscoder(nil)

      end

      def refreshViewers()
        @findings_tree.reload()
        @sites_tree.reload()
        #@chatTable.clearItems()
      end

      def onOpenInterceptor(sender, sel, ptr)
        if @active_project then
          interceptor = Watobo::Core::FXInterceptor.new(self, @iproxy, :opts => DECOR_ALL)
          iproxy.target = interceptor
          puts "* Interceptor created"
          #@active_project.interceptor = interceptor
          interceptor.create
          interceptor.show(Fox::PLACEMENT_SCREEN)
          getApp().runModalWhileShown(interceptor)
          interceptor.releaseAll()
          puts "* Interceptor closed"
          iproxy.target = nil
          #if interceptor.execute != 0 then
          #  puts "interceptor finished"
          #end
        else

          FXMessageBox.information(self,MBOX_OK,"No Project Defined", "Create Project First")
        end
      end

      def setStatus(project)
        @statusBar.projectName = project.projectName
        @statusBar.sessionName = project.sessionName
        @statusBar.portNumber = @settings[:interceptor_prefs][:port].to_s
        @statusBar.forwardingProxy = "-"
        proxy = project.getCurrentProxy()
        unless proxy.nil?

          @statusBar.forwardingProxy = "#{proxy[:name]} (#{proxy[:host]}:#{proxy[:port]})"
        end
        @statusBar.statusInfo = "Ready"
        @scan_button.enable
        @dashboard.updateProjectInfo(project)
      end

      def setupProgressWindow(title, numTotal)
        @progressWindow = Watobo::Gui::ProgressWindow.new(self, title, numTotal)

        @progressWindow.create
        @progressWindow.show(PLACEMENT_SCREEN)
        #  getApp().runModalWhileShown(@progressWindow)
        return @progressWindow

      end

      def closeProgressWindow
        # getApp().stopModal()
        @progressWindow.finished
      end

      def disableFileMenu
        @file_new_menu.enabled = false
        @file_history.disable
        @new_project_button.disable
      end

      def closeProject()
        @active_project = nil
        @findings_tree.clearItems()
        @sites_tree.clearItems()
        @chatTable.clearItems()
        @request_viewer.setText('')
        @response_viewer.setText('')
        @lastViewed = nil
        @last_request = nil
        @last_response = nil
        @iproxy.stop if @iproxy

      end

      #
      # onNewProject
      #
      def onNewProject(sender,sel,ptr)

        if @active_project then
          response = FXMessageBox.question(self, MBOX_YES_NO, "New Project", "This will close the actual project!\nAre you sure?")
          return 0 if not response == MBOX_CLICKED_YES
          # clear old project
          closeProject()
          # stop interceptor
        end

        newProjectWizzard = Watobo::Gui::NewProjectWizzard.new(self, @settings[:workspace_path])
        if newProjectWizzard.execute != 0
          # prepare project settings
          new_project_settings = {
            :project_path => newProjectWizzard.selected_project_path,
            :session_path => newProjectWizzard.selected_session_path,
            :project_name => newProjectWizzard.project_name,
            :session_name => newProjectWizzard.session_name
          }
          @settings[:workspace_path] = newProjectWizzard.workspace_dir

          new_project_settings.update(@settings)

          file = new_project_settings[:project_name] + @settings[:project_settings_file_ext]
          project_settings_file = File.join(new_project_settings[:project_path], file)
          if File.exist?(project_settings_file) then
            puts "* load project settings: #{project_settings_file}"
            new_project_settings.update loadProjectSettings(project_settings_file)
          end
          # setup project settings filename -> history
          file = new_project_settings[:session_name] + @settings[:session_settings_file_ext]
          session_settings_file = File.join(new_project_settings[:session_path], file)

          if File.exist?(session_settings_file) then
            puts "* load session settings: #{session_settings_file}"
            new_project_settings.update loadSessionSettings(session_settings_file)
          end

          @settings[:history].unshift  session_settings_file if not @settings[:history].include?(session_settings_file)
          @settings[:history].pop if @settings[:history].length > 5

          saveDefaultSettings()

          puts "creating new project ..."
          print "Name: "
          print "#{new_project_settings[:project_name]}\n\n"

          project = Watobo::Core::Project.new(new_project_settings)

          if project
            @projects[project] = {
              :session_file => session_settings_file,
              :project_file => project_settings_file
            }
          end

          startProject(project)

        end

      end

      def startProject(project)
        begin
          progressWindow = Watobo::Gui::ProgressWindow.new(self, "New Project", 0)

          progressWindow.create
          progressWindow.show(PLACEMENT_SCREEN)

          if project then
            @active_project = project
            saveProjectSettings(@active_project)
            saveSessionSettings(@active_project)

            subscribeProject(@active_project)

            Thread.new(progressWindow){ |pw|
              @chatTable.hide
              @sites_tree.hide
              @findings_tree.hide
              @active_project.setupProject(pw)
              @chatTable.show
              @sites_tree.show
              @sites_tree.recalc
              #@sites_tree.handle(self, FXSEL(SEL_CHANGED, 0), nil)
              @findings_tree.show
              @findings_tree.recalc

              @sites_tree.project = @active_project
              @findings_tree.project = @active_project
              pw.hide
            }

            getApp().runModalWhileShown(progressWindow)

            @iproxy = Watobo::Core::InterceptProxy.new(@settings[:interceptor_prefs])
            @iproxy.subscribe(:new_interception) { |chat|
              Thread.new(chat) { |c|
                @active_project.addChat(c)
              }
            }
            @iproxy.run()
            @browserView = BrowserPreview.new(@iproxy)

            updateRequestTable(@active_project)

            setStatus(@active_project)

          else
            puts "!!! Could not create project :("
          end
        rescue => bang
          puts "!!! Could not create project"
          puts bang
          puts bang.backtrace
        end
      end

      def openSession(session_file)

        if @active_project then
          response = FXMessageBox.question(self, MBOX_YES_NO, "New Project", "This will close the actual project!\nAre you sure?")
          return false if not response == MBOX_CLICKED_YES
          # clear old project
          closeProject()
          # stop interceptor
        end

        session_settings = nil
        if File.exists?(session_file) then
          session_settings = Watobo::Utils.loadSettings(session_file)

          puts "== Loaded Session Settings ==="
          puts session_settings.to_yaml
          puts "-------------------------------"
          #  gets
        else
          puts "!!! Session file does not exist (#{session_file})."
          return false
        end

        return false if !session_settings[:project_name]

        project_file = File.join(File.dirname(session_file), "..", session_settings[:project_name] + @settings[:project_settings_file_ext])
        if File.exists?(project_file) then
          project_settings = Watobo::Utils.loadSettings(project_file)
          if not project_settings.is_a? Hash then
            project_settings = Hash.new
          end
          new_project_settings = Hash.new
          new_project_settings.update(@settings)
          new_project_settings.update(project_settings)
          new_project_settings.update(session_settings)
          new_project_settings[:workspace_path] = File.join(File.dirname(project_file), "..")
          #       puts "== Loaded Project Settings ==="
          #     puts project_settings.to_yaml
          #     puts "-------------------------------"
          #     gets
          project = Watobo::Core::Project.new(new_project_settings)

          if project
            @projects[project] = {
              :session_file => session_file,
              :project_file => project_file
            }
            startProject(project)
          end

          return true
        else
          puts "!!! No project file available (#{project_file})."
          return nil
        end

        #  return 1
      end

      def setActiveProject(project)
        @active_project = project
      end

      def openScannerSettingsDialog(sender,sel,ptr)
        if @active_project then
          settings = @active_project.getScanPreferences()
          # puts settings.to_yaml
          dlg = Watobo::Gui::ScannerSettingsDialog.new(self, settings, LAYOUT_FILL_X|LAYOUT_FILL_Y)
          if dlg.execute != 0 then
            puts dlg.scanner_settings.to_yaml
            @active_project.updateSettings(YAML.load(YAML.dump(dlg.scanner_settings)))
            saveProjectSettings(@active_project)

          end
        else
          FXMessageBox.information(self,MBOX_OK,"No Project Defined", "Create Project First")
        end
      end

      def openInterceptorSettingsDialog(sender,sel,ptr)
        dlg = Watobo::Gui::InterceptorSettingsDialog.new(self, @settings[:interceptor_prefs], LAYOUT_FILL_X|LAYOUT_FILL_Y)
        if dlg.execute != 0 then
          puts dlg.interceptor_settings.to_yaml
          @settings[:interceptor_prefs].update YAML.load(YAML.dump(dlg.interceptor_settings))
          #@active_project.updateSettings(YAML.load(YAML.dump(dlg.scanner_settings)))
          FXMessageBox.information(self,MBOX_OK,"Restart required?", "You must restart your project in order your changes take effect.")
          saveDefaultSettings(@settings[:interceptor_prefs])
        end
      end

      def openScopeDialog(sender,sel,ptr)
        dlg = Watobo::Gui::EditScopeDialog.new(self, @active_project, LAYOUT_FILL_X|LAYOUT_FILL_Y)
        if dlg.execute != 0 then
          @active_project.scope = YAML.load(YAML.dump(dlg.scope))
          saveProjectSettings(@active_project)
          refreshViewers()
        end
      end

      def startFullScan(sender,sel,ptr)
        if @scan_button.icon == ICON_STOP
          @scanner.cancel() if @scanner
          @scan_button.icon = ICON_START
          @scan_running = false
        else
          dlg = Watobo::Gui::FullScanDialog.new(self, @active_project, LAYOUT_FILL_X|LAYOUT_FILL_Y)
          if dlg.execute != 0 then

            @scan_running = true
            @scan_button.icon = ICON_STOP
            #@pause_scan_button.enable
            @active_project.setScope(dlg.scope)

            saveProjectSettings(@active_project)
            saveSessionSettings(@active_project)

            @active_project.scanLogDirectory = dlg.scan_prefs[:scanlog_dir] unless dlg.scan_prefs[:scanlog_dir] == ''

            #   chats = @active_project.getChats(dlg.scope)
            active_modules = dlg.activeModules()
            acc = @active_project.listActiveCheckClasses

            acc.map! {|x|
              r = nil
              active_modules.each do |am|
                r = x if x.new(:dummy).info[:check_name] == am.info[:check_name]
              end
              r
            }
            acc.compact!

            in_scope_chats = @active_project.chatsInScope()
            puts "Chats to scan:"
            puts in_scope_chats.length
            p

            confirm_dlg = Watobo::Gui::ConfirmScanDialog.new(self, in_scope_chats, @active_project.getScanPreferences())
            if confirm_dlg.execute == 0
              @scan_button.icon = ICON_START
              @scan_running = false
              return 0
            end

            scan_prefs = @active_project.getScanPreferences()

            scan_prefs[:scanlog_dir] = dlg.scan_prefs[:scanlog_dir]
            if File.exist?(scan_prefs[:scanlog_dir]) then
              new_dir = File.join(scan_prefs[:scanlog_dir], "scan-" + Time.now.to_f.to_s + "_full")
              Dir.mkdir(new_dir)
              scan_prefs[:scanlog_dir] = new_dir
            end

            @scanner = Watobo::Core::Scanner2.new(in_scope_chats, acc , @active_project.passive_checks, scan_prefs)

            @scanner.subscribe(:progress) { |check|
              Thread.new(check) { |cc|
                begin
                  @dashboard.progress(cc)
                rescue => bang
                  p bang
                  p bang.backtrace
                end
              }
            }

            @scanner.subscribe(:logger){ |level, message|
              logger(message)
            }

            @scanner.subscribe(:new_finding) { |finding|
              Thread.new(finding) { |f|
                begin
                  @active_project.addFinding(f)
                rescue => bang
                  p bang
                  p bang.backtrace
                end
              }
            }
            # p @scanner.numChecksPerModule.to_yaml
            #gets
            @dashboard.setupScanProgressFrame(@scanner.numChecksPerModule)

            Thread.new {
              begin
                @dashboard.setScanStatus("Running")
                @statusBar.statusInfo = "Full Scan Running"
                @scanner.run(:run_passive_checks => false, :update_sids => true, :update_session => true)
                logger("scan finished")
                @dashboard.setScanStatus("Finished")
                @statusBar.statusInfo = "Ready."
                @scan_button.icon = ICON_START
                @scan_running = false
              rescue => bang
                puts bang
              end
            }
          end
        end
      end

      def pauseScan(sender, sel, ptr)
        begin
          if @scanner.running?
            @scanner.stop
            @dashboard.setScanStatus("Scan Paused")
            @statusBar.statusInfo = "Scan Paused"
          else
            @scanner.continue
            @dashboard.setScanStatus("Scan Running")
            @statusBar.statusInfo = "Full Scan Running"
          end
        rescue => bang
          puts "!!!ERROR: Could not pause scanner"
          puts bang
        end
      end

      def applyFilter
        if @active_project then
          #  @doctype_TableFilter=""
          @doctype_TableFilter.clear
          if @foption_nopix.checked? then @doctype_TableFilter.concat(@fext_pix);end
          if @foption_nodocs.checked? then @doctype_TableFilter.concat(@fext_docs);end
          if @foption_nojs.checked? then @doctype_TableFilter.concat(@fext_javascript);end
          if @foption_nocss.checked? then @doctype_TableFilter.concat(@fext_style);end

          updateRequestTable(@active_project)
        end
      end

      #
      # new_event
      #
      def logger(message)
        t = Time.now
        now = t.strftime("%m/%d/%Y @ %H:%M:%S")
        @logText.appendText("#{now}: #{message}\n")

      end

      def init_main_UNUSED(qmain)
        @main=qmain
      end

      def initialize(app)
        # Invoke base class initialize first
        super(app, "WATOBO by siberas (Version: #{$version})", :opts => DECOR_ALL, :width => 1000, :height => 600)
        # FXToolTip.new(app)

        self.icon = ICON_WATOBO
        self.show(PLACEMENT_MAXIMIZED)
        self.connect(SEL_CLOSE, method(:onClose))

        @projects = Hash.new

        @scanner = nil
        @iproxy = nil
        @browserView = nil

        @scan_running = false

        @lastViewed = nil # last viewed/selected item (chat/finding)
        @last_request = nil
        @last_response = nil

        @settings = {
          :project_settings_file_ext => '.wps',
          :session_settings_file_ext => '.wss',
          :policy => { :name => 'Default',
          :list => {
          'Default' => {
          'Watobo::Modules::Active::Directories::Dir_indexing' => false,
          'Watobo::Modules::Active::Discovery::Http_methods' => false,
          'Watobo::Modules::Active::Domino::Domino_db' => false,
          'Watobo::Modules::Active::Sap::Its_commands' => false,
          'Watobo::Modules::Active::Sap::Its_services' => false,
          'Watobo::Modules::Active::Sap::Its_service_parameter' => false,
          'Watobo::Modules::Active::Sap::Its_xss' => false,
          'Watobo::Modules::Active::Sqlinjection::Sqli_simple' => true,
          'Watobo::Modules::Active::Sqlinjection::Sql_boolean' => true,
          'Watobo::Modules::Active::Sqlinjection::Sql_numerical' => false,
          'Watobo::Modules::Active::Xss::Xss_simple' => true,
          }
          }
          },
          :sid_patterns => [
          "name=\"(sessid)\" value=\"([0-9a-zA-Z!-]*)\"",
          "(sessid)=([-0-9a-zA-Z_:]*)(;|&)?",
          '(SESSIONID)=([-0-9a-zA-Z_:\.\(\)]*)(;|&)?',
          "(PHPSESSID)=([0-9a-zA-Z]*)(;|&)?",
          '(ASPSESSIONID)\w*=([0-9a-zA-Z]*)(;|&)?',
          "(MYSAPSSO2)=([0-9a-zA-Z.=%]*)(;|&)?",
          "(ELEXIRSID)=([0-9a-zA-Z!-]*)(;|&)?",
          "(SLSID)=([0-9a-zA-Z!-]*)(;|&)?",
          "(sid)=([0-9a-z]*)(')?", #servlet?sid=9912ad967cc578a12ada85d91f841e18')
          '(saplb_\*)=([-0-9a-zA-Z_:\(\)]*)(;|&)?',
          "(DomAuthSessId)=([0-9a-zA-Z]*)(;|&)?",
          '(wgate)\/([\w]{4,}\/[\w=~]*)(;|&|\'|")?', # SAP ITS WGATE Session Handling
          '(session)=([-0-9a-zA-Z_:\.]*)(;|&)?'       # SAP ITS Session Cookie
          ],
          :logout_signatures => [
          "^Location.*login"
          ],
          :forwarding_proxy => {
          :name => nil,
          :list => {
          'webscarab' => {
          :host => '127.0.0.1',
          :port => "8008",
          :name => "webscarab",
          :description => "",
          :auth_type => "",
          :auth_user => "",
          :auth_password => "",
          :auth_domain => "",
          :auth_workstation => ""
          }

          } # e.g., list[watobo] = { :proxy => name|ip, :port => port, :description, :auth_type }
          },
          :interceptor_prefs => {
          :port => 8081,

          :pass_through => {
          :content_types => [ 'application\/audio',
          'application\/video',
          'application\/image',
          'application\/pdf',
          'application\/.*flash',
          'image\/'
          ],
          :content_length => 100000
          },

          :certificate_path => "certificates",
          :cert_file => "cert.pem",
          :key_file => "privkey.pem"

          },

          :history => [],
          :workspace_path => '',
          :smart_scan => true,

        }

        @default_settings_file = File.join($watobo_base, "watobo.wds")
        print "* loading default settings from file:\r\n>> #{@default_settings_file} ... "
        if File.exists?(@default_settings_file) then
          print "[OK]\n"
          @settings.update(Watobo::Utils.loadSettings(@default_settings_file))
          # puts @default_settings.to_yaml
          # gets
        else
          print "[FALSE]\n"
        end

        @app = app

        @progressWindow = nil             # reserved for simple progress Window

        @switcher = nil

        @interceptor = nil

        @doctype_TableFilter = []
        #--------------------------------
        # file extensions
        #--------------------------------
        @fext_pix = %w[ jpg gif bmp ico png jpeg ]
        @fext_docs = %w[ ppt doc xls pptx docx xlsx pdf ]
        @fext_javascript = %w[ js json ]
        @fext_style = %w[ css ]

        @table_filter = FXDataTarget.new("")
        @active_project = nil

        # setup clipboard
        @clipboard_text = ""
        self.connect(SEL_CLIPBOARD_REQUEST) do
          setDNDData(FROM_CLIPBOARD, FXWindow.stringType, Fox.fxencodeStringData(@clipboard_text))
        end

        menu_bar = FXMenuBar.new(self, :opts => LAYOUT_SIDE_TOP|LAYOUT_FILL_X)

        file_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "File" , :popupMenu => file_menu_pane)
        @file_new_menu = FXMenuCommand.new(file_menu_pane, "New/Open" )
        @file_new_menu.connect(SEL_COMMAND, method(:onNewProject))

        exit_menu = FXMenuCommand.new(file_menu_pane, "Exit" )
        #FXMenuCommand.new(file_menu_pane, "Exit", nil, getApp(), FXApp::ID_QUIT)
        exit_menu.connect(SEL_COMMAND, method(:onExit))

        FXMenuSeparator.new(file_menu_pane)

        submenu = FXMenuPane.new(self) do |session_menu|

          @settings[:history].each do |h|
            # puts "* HISTORY #{h}"
            if File.exists?(h)
              # ws = File.expand_path(File.join(h,"..",".."))
              h.gsub!(/\\/, "/")
              hname = h
              if h =~ /(\/[^\/]*\/[^\/]*\/[^\/]*)\/[^\/]*$/
                hname = "..." + $1
              end

              history = FXMenuCommand.new(session_menu, hname )
              history.connect(SEL_COMMAND) do |sender, sel, item|
                puts "open session #{h}"
                puts "!!!ERROR Could not start session #{h}" if !openSession(h)

              end
            end
          end
        end
        @file_history = FXMenuCascade.new(file_menu_pane, "Recent Sessions", nil, submenu)

        #  file_open_command = FXMenuCommand.new(file_menu_pane, "Open..." )
        #  file_save_command = FXMenuCommand.new(file_menu_pane, "Save" )
        #  file_save_as_command = FXMenuCommand.new(file_menu_pane, "Save As..." )

        settings_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "Settings" , :popupMenu => settings_menu_pane)
        menu_proxy = FXMenuCommand.new(settings_menu_pane, "Forwarding Proxy..." )
        menu_session = FXMenuCommand.new(settings_menu_pane, "Session Management..." )
        # @active_project ? menu_session.enable : menu_session.disable

        menu_ca = FXMenuCommand.new(settings_menu_pane, "Create Certificate..." )
        menu_ca.connect(SEL_COMMAND, method(:openCADialog))

        menu_preferences = FXMenuCommand.new(settings_menu_pane, "Target Scope..." )
        menu_preferences.connect(SEL_COMMAND, method(:openScopeDialog))

        menu_preferences = FXMenuCommand.new(settings_menu_pane, "Scanner..." )
        menu_preferences.connect(SEL_COMMAND, method(:openScannerSettingsDialog))

        menu_preferences = FXMenuCommand.new(settings_menu_pane, "Interceptor..." )
        menu_preferences.connect(SEL_COMMAND, method(:openInterceptorSettingsDialog))

        www_auth_prefs = FXMenuCommand.new(settings_menu_pane, "WWW-Auth..." )
        www_auth_prefs.connect(SEL_COMMAND) { openWwwAuthDialog() }
        # intercept_enable = FXMenuCheck.new(settings_menu_pane, "Enable Interception")

        # file_menu_title = FXMenuTitle.new(menu_bar, "Settings" , :popupMenu => settings_menu_pane)

        menu_proxy.connect(SEL_COMMAND, method(:onMenuProxy))
        menu_session.connect(SEL_COMMAND, method(:openSessionManagement))

        tools_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "Tools" , :popupMenu => tools_menu_pane)
        transcoder = FXMenuCommand.new(tools_menu_pane, "Transcoder")
        interceptor = FXMenuCommand.new(tools_menu_pane, "Interceptor")
        transcoder.connect(SEL_COMMAND, method(:onOpenTranscoder))
        interceptor.connect(SEL_COMMAND, method(:onOpenInterceptor))

        view_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "View" , :popupMenu => view_menu_pane)
        view_logs_command = FXMenuCommand.new(view_menu_pane, "Logs" )
        view_dashboard_command = FXMenuCommand.new(view_menu_pane, "Dashboard" )
        view_findings_command = FXMenuCommand.new(view_menu_pane, "Chat-Table")

        view_dashboard_command.connect(SEL_COMMAND, method(:showDashboard))
        view_logs_command.connect(SEL_COMMAND, method(:showLogs))
        view_findings_command.connect(SEL_COMMAND, method(:showConversation))

        window_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "Window" , :popupMenu => window_menu_pane)
        use_small_icons = FXMenuCheck.new(window_menu_pane, "Small Icons/Text" )
        use_small_icons.connect(SEL_COMMAND) {
          if use_small_icons.checked?
            useSmallIcons()
          else
            useRegularIcons()
          end
        }

        help_menu_pane = FXMenuPane.new(self)
        FXMenuTitle.new(menu_bar, "Help" , :popupMenu => help_menu_pane)
        #   menu_lic = FXMenuCommand.new(help_menu_pane, "License" )
        menu_about = FXMenuCommand.new(help_menu_pane, "About" )
        menu_about.connect(SEL_COMMAND) {
          #FXMessageBox.information(self,MBOX_OK,"About", "WATOBO Version 0.9.1!")
          aboutDlg = AboutWatobo.new(self)
          aboutDlg.create
          aboutDlg.show(Fox::PLACEMENT_SCREEN)
        }

        #   top_dock_frame = FXVerticalFrame.new(self, :opts => LAYOUT_FILL_X|FRAME_SUNKEN, :padding => 0)
        #   tool_bar_shell = FXToolBarShell.new(self)
        #   top_dock_site = FXDockSite.new(top_dock_frame, :opts => LAYOUT_FILL_X|LAYOUT_SIDE_TOP)
        # bottom_dock_site = FXDockSite.new(self, :opts => LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM)

        #  project_bar = FXToolBar.new(top_dock_site, tool_bar_shell, PACK_UNIFORM_WIDTH)
        #  FXToolBarGrip.new(project_bar, :opts => TOOLBARGRIP_SINGLE)
        top_bar = FXHorizontalFrame.new(self, :opts => LAYOUT_FILL_X|FRAME_SUNKEN, :padding => 0)
        project_bar = FXHorizontalFrame.new(top_bar, :opts => LAYOUT_FILL_X|PACK_UNIFORM_WIDTH|FRAME_NONE, :padding => 2)
        @new_project_button = FXButton.new(project_bar, "\tNew Project\tNew Project." , :icon => ICON_ADD_PROJECT, :padding => 0)
        @new_project_button.connect(SEL_COMMAND, method(:onNewProject))

        @scan_button = FXButton.new(project_bar, "\tStart Scan\tStart Scan." ,:opts => FRAME_RAISED|FRAME_THICK, :icon => ICON_START, :padding => 0)
        @scan_button.disable
        # @start_scan_button.tipText = "Start Scan"
        @scan_button.connect(SEL_COMMAND, method(:startFullScan))

        #@pause_scan_button = FXButton.new(project_bar, "\tPause Scan\tPause Scan." ,:icon => ICON_PAUSE)
        #@pause_scan_button.disable
        # @start_scan_button.tipText = "Start Scan"
        #@pause_scan_button.connect(SEL_COMMAND, method(:pauseScan))

        #   views_bar = FXToolBar.new(top_dock_site, tool_bar_shell, :opts => PACK_UNIFORM_WIDTH, :height => 20)
        #   FXToolBarGrip.new(views_bar, :opts => TOOLBARGRIP_SINGLE)

        #@create_report_button = FXButton.new(views_bar,"\tReport\tCreate Report." , :icon => @reportIcon)
        #@create_report_button.disable
        views_bar = project_bar
        @show_dashboard_button = FXButton.new(views_bar, "\tShow Dashboard\tShow Dashboard.", :icon => ICON_DASHBOARD, :padding => 0)
        @show_dashboard_button.connect(SEL_COMMAND, method(:showDashboard))
        # @create_report_button.disable
        @show_conversation_button = FXButton.new(views_bar, "\tShow Conversation\tShow Conversation.", :icon => ICON_CONVERSATION, :padding => 0)
        @show_conversation_button.connect(SEL_COMMAND, method(:showConversation))

        #  tools_bar = FXToolBar.new(top_dock_site, tool_bar_shell, :opts => PACK_UNIFORM_WIDTH)
        #  FXToolBarGrip.new(tools_bar, :opts => TOOLBARGRIP_SINGLE)
        tools_bar = project_bar
        @open_transcoder_button = FXButton.new(tools_bar, "\tOpen Transcoder\tOpen Transcoder.", :icon => ICON_TRANSCODER, :padding => 0)
        @open_transcoder_button.connect(SEL_COMMAND, method(:onOpenTranscoder))

        @btn_show_plugin = FXButton.new(tools_bar, "\tShow Plugins\tShow Plugins.", :icon => ICON_PLUGIN, :padding => 0)
        @btn_show_plugin.connect(SEL_COMMAND, method(:onShowPlugins))

        FXButton.new(top_bar, nil, :icon => SIBERAS_ICON, :opts => FRAME_NONE|LAYOUT_SIDE_RIGHT)

        @statusBar = StatusBar.new(self, :opts => LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM, :padding => 0)

        splitter = FXSplitter.new(self, :opts => LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_TRACKING)

        frame = FXVerticalFrame.new(splitter, :opts => LAYOUT_FILL_Y|LAYOUT_FIX_WIDTH|FRAME_SUNKEN, :padding => 0, :width => 200)

        @treeTabbook = FXTabBook.new(frame, nil, 0, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_RIGHT)

        ftab = FXTabItem.new(@treeTabbook, "Findings", nil)
        ftab.setFont(FXFont.new(getApp(), "helvetica", 12, FONTWEIGHT_BOLD, FONTSLANT_ITALIC, FONTENCODING_DEFAULT))
        tab_frame = FXVerticalFrame.new(@treeTabbook, :opts => LAYOUT_FIX_WIDTH|LAYOUT_FILL_Y|FRAME_RAISED, :width => 100)
        frame = FXVerticalFrame.new(tab_frame, :opts => LAYOUT_FILL_Y|LAYOUT_FILL_X|FRAME_SUNKEN, :padding => 0)
        @findings_tree = Watobo::Gui::FindingsTree.new(frame, self, nil)
        stab = FXTabItem.new(@treeTabbook, "  Sites  ", nil)
        stab.setFont(FXFont.new(getApp(), "helvetica", 12, FONTWEIGHT_BOLD, FONTSLANT_ITALIC, FONTENCODING_DEFAULT))
        tab_frame = FXVerticalFrame.new(@treeTabbook, :opts => LAYOUT_FIX_WIDTH|LAYOUT_FILL_Y|FRAME_RAISED, :width => 100)
        frame = FXVerticalFrame.new(tab_frame, :opts => LAYOUT_FILL_Y|LAYOUT_FILL_X|FRAME_SUNKEN, :padding => 0)
        @sites_tree = Watobo::Gui::SitesTree.new(frame, self, nil)

        #   @treeTabbook.connect(SEL_COMMAND) { |sender, sel, item|
        #     case item
        #       when 0
        # if @active_project
        #   @active_project.settings.delete(:site_filter)
        #       updateRequestTable(@active_project)
        #end
        #     end
        #   }
        subscribeFindingsTree()
        subscribeSitesTree()
        #
        # S W I T C H E R
        #

        @switcher = FXSwitcher.new(splitter,LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)

        # R E Q U E S T I N F O
        requestInfo = FXVerticalFrame.new(@switcher, :opts => LAYOUT_FILL_X|LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        request_splitter = FXSplitter.new(requestInfo, :opts => LAYOUT_SIDE_TOP|SPLITTER_HORIZONTAL|LAYOUT_FILL_Y|LAYOUT_FILL_X|SPLITTER_TRACKING|SPLITTER_REVERSED)

        conversation_frame = FXVerticalFrame.new(request_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN)

        # filterFrame =FXVerticalFrame.new(conversation_frame, LAYOUT_FILL_X)

        @tabBook = FXTabBook.new(conversation_frame, nil, 0, :opts => LAYOUT_FILL_X|LAYOUT_RIGHT, :padding => 0)

        docfilter_tab = FXTabItem.new(@tabBook, "Doc Filter", nil)

        docfilter_frame = FXHorizontalFrame.new(@tabBook, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED)

        @foption_nopix = FXCheckButton.new(docfilter_frame, "excl. pics", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @foption_nopix.setCheck(true)
        @foption_nodocs = FXCheckButton.new(docfilter_frame, "excl. docs", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @foption_nodocs.setCheck(true)
        @foption_nojs = FXCheckButton.new(docfilter_frame, "excl. javascript", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @foption_nojs.setCheck(true)
        @foption_nocss = FXCheckButton.new(docfilter_frame, "excl. style sheets", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @foption_nocss.setCheck(true)

        search_tab = FXTabItem.new(@tabBook, "Text Filter", nil)

        search_frame = FXHorizontalFrame.new(@tabBook, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED)
        FXButton.new(search_frame, "Clear", nil, nil, 0, FRAME_RAISED|FRAME_THICK).connect(SEL_COMMAND, method(:onClear))
        @tableFilterFX = FXTextField.new(search_frame, 20, @table_filter, FXDataTarget::ID_VALUE, FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X)
        # filterOptionsFrame =FXHorizontalFrame.new(fbox, LAYOUT_FILL_X)
        @foption_req = FXCheckButton.new(search_frame, "Full Request", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @foption_res = FXCheckButton.new(search_frame, "Full Response", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)

        options_tab = FXTabItem.new(@tabBook, "Options", nil)
        options_frame = FXHorizontalFrame.new(@tabBook, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED)
        @table_option_autoscroll = FXCheckButton.new(options_frame, "autoscroll", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @table_option_autoscroll.setCheck(true)

        @table_option_unique = FXCheckButton.new(options_frame, "unique chats", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @table_option_unique.setCheck(false)

        @table_option_scope = FXCheckButton.new(options_frame, "scope only", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @table_option_scope.setCheck(false)

        @table_option_hidetested_cb = FXCheckButton.new(options_frame, "hide tested", nil, 0, ICON_BEFORE_TEXT|LAYOUT_SIDE_LEFT)
        @table_option_hidetested_cb.setCheck(false)
        #applyFilterButton = FXButton.new(conversation_frame, "Apply", nil, nil, 0, FRAME_RAISED|FRAME_THICK)
        docfilter_button_frame = FXHorizontalFrame.new(conversation_frame, :opts => LAYOUT_FILL_X, :padding => 0)
        FXButton.new(docfilter_button_frame, "Apply", nil, nil, 0, FRAME_RAISED|FRAME_THICK).connect(SEL_COMMAND, method(:onApplyFilterClick))
        #FXButton.new(docfilter_button_frame, "Clear", nil, nil, 0, FRAME_RAISED|FRAME_THICK).connect(SEL_COMMAND, method(:onClear))

        @table_filter.connect(SEL_COMMAND) do |sender, sel, event|
          # puts "apply filter #{@table_filter.value}"
          applyFilter()
        end

        FXButton.new(docfilter_button_frame, "", ICON_BTN_UP, nil, 0, FRAME_RAISED|FRAME_THICK).connect(SEL_COMMAND) {
          @chatTable.scrollUp()
        }

        FXButton.new(docfilter_button_frame, "", ICON_BTN_DOWN, nil, 0, FRAME_RAISED|FRAME_THICK).connect(SEL_COMMAND) {
          @chatTable.scrollDown()
        }

        #==============================================================================================#
        # Chat-Table
        #==============================================================================================#

        #@chatTable = FXTable.new(conversation_frame,
        @chatTable = ConversationTable.new(conversation_frame,
        :opts => TABLE_COL_SIZABLE|TABLE_ROW_SIZABLE|LAYOUT_FILL_X|LAYOUT_FILL_Y|TABLE_READONLY,
        :padding => 2)

        @table_option_autoscroll.connect(SEL_COMMAND) {
          # puts @table_option_autoscroll.checked?
          if @table_option_autoscroll.checked? == true
            @chatTable.autoscroll =  true
          else
            @chatTable.autoscroll = false
          end
        }

        @chatTable.autoscroll =  true

        @chatTable.connect(SEL_COMMAND) do |sender, sel, item|
          @findings_tree.killSelection()
          @sites_tree.killSelection()
          onTableClick(sender,sel,item)
        end

        @chatTable.connect(SEL_DOUBLECLICKED) do |sender, sel, data|
          @findings_tree.killSelection()
          @sites_tree.killSelection()
          row = sender.getCurrentRow
          if row >= 0 then
            @chatTable.selectRow(row, false)
            chatid = @chatTable.getRowText(row).to_i
            chat = @active_project.getChat(chatid)
            open_manual_request_toolkit(chat)
          end
        end

        @chatTable.connect(SEL_RIGHTBUTTONRELEASE) do |sender, sel, event|
          @findings_tree.killSelection()
          @sites_tree.killSelection()
          unless event.moved?
            #   row = sender.getCurrentRow
            ypos = event.click_y
            row = @chatTable.rowAtY(ypos)
            #  puts "right click on row #{row} of #{@chatTable.numRows}"
            if row >= 0 and row < @chatTable.numRows then

              @chatTable.selectRow(row, false)
              chatid = @chatTable.getRowText(row).to_i
              chat = @active_project.getChat(chatid)

              showChat(chat)

              FXMenuPane.new(self) do |menu_pane|

                # SEND TO SUBMENU
                submenu = FXMenuPane.new(self) do |sendto_menu|

                  target = FXMenuCommand.new(sendto_menu, "Fuzzer..." )
                  target.connect(SEL_COMMAND) {
                    openFuzzer(chat)
                  }
                  target = FXMenuCommand.new(sendto_menu, "Manual Request..." )
                  target.connect(SEL_COMMAND) {
                    open_manual_request_toolkit(chat)
                  }

                end
                FXMenuCascade.new(menu_pane, "Send to", nil, submenu)

                # EXCLUDE SUBMENU
                exclude_submenu = FXMenuPane.new(self) do |sub|
                  chat = @active_project.getChat(chatid)

                  target = FXMenuCheck.new(sub, "Chat (#{chatid})" )

                  target.check = @active_project.settings[:excluded_chats].include?(chatid) ? true : false

                  target.connect(SEL_COMMAND) { |sender, sel, item|
                    if sender.checked?()
                      @active_project.settings[:excluded_chats].push chatid
                    else
                      @active_project.settings[:excluded_chats].delete(chatid)
                    end

                  }
                  #   target = FXMenuCommand.new(sub, "Path" )
                  #   target.connect(SEL_COMMAND) {
                  # ...
                  #   }

                end
                FXMenuCascade.new(menu_pane, "Exclude from Scan", nil, exclude_submenu)

                # COPY SUBMENU
                copy_submenu = FXMenuPane.new(self) do |sub|
                  chat = @active_project.getChat(chatid)
                  url = @active_project.getChat(chatid).request.url
                  url_string = "URL: #{url.slice(0,35)}"
                  url_string += "..." if url.length > 36
                  target = FXMenuCommand.new(sub, url_string )
                  target.connect(SEL_COMMAND) {
                    types = [ FXWindow.stringType ]
                    if acquireClipboard(types)
                      @clipboard_text = url
                    end

                  }
                  target = FXMenuCommand.new(sub, "Site: #{@active_project.getChat(chatid).request.site}" )
                  target.connect(SEL_COMMAND) {
                    url = @active_project.getChat(chatid).request.site

                    types = [ FXWindow.stringType ]
                    if acquireClipboard(types)
                      @clipboard_text = url
                    end
                  }

                end
                FXMenuCascade.new(menu_pane, "Copy", nil, copy_submenu)

                addToLogin = FXMenuCommand.new(menu_pane, "Add to Login-Script" )
                addToLogin.connect(SEL_COMMAND) {
                  @active_project.settings[:login_chat_ids].push chatid
                  saveSessionSettings(@active_project)
                }

                target = FXMenuCheck.new(menu_pane, "Mark as tested" )
                target.check = chat.tested?
                target.connect(SEL_COMMAND) { |sender, sel, item|
                  chat.tested = sender.checked?()
                  Watobo::Utils.saveChat(chat, chat.file)
                }

                FXMenuCommand.new(menu_pane, "Edit comment.." ).connect(SEL_COMMAND) {
                  dlg = Watobo::Gui::EditCommentDialog.new(self, chat)
                  if dlg.execute != 0 then
                    chat.comment = dlg.comment
                    Watobo::Utils.saveChat(chat, chat.file)
                  end
                }
                #  copyRequest = FXMenuCommand.new(menu_pane, "copy Request(#{chatid})" )
                #  copyResponse = FXMenuCommand.new(menu_pane, "copy Response(#{chatid})" )

                #  info = FXMenuCommand.new(menu_pane, "Details..." )
                #info.connect(SEL_COMMAND) { display_info_for(item) }

                menu_pane.create
                menu_pane.popup(nil, event.root_x, event.root_y)
                app.runModalWhileShown(menu_pane)

              end

            end

          end

        end

        #===================================================================
        # CHAT VIEWER
        #===================================================================
        chat_frame = FXVerticalFrame.new(request_splitter, :opts => LAYOUT_FILL_Y|LAYOUT_FILL_X|FRAME_SUNKEN|LAYOUT_MIN_WIDTH, :padding => 0, :width=>400)

        #view_menu = FXVerticalFrame.new(chat_frame, :opts => LAYOUT_FILL_X, :padding => 0)
        #frame = FXHorizontalFrame.new(view_menu, :opts => LAYOUT_FILL_X)
        #cvlabel = FXLabel.new(frame, "View: ")
        # cvlabel.setFont(FXFont.new(getApp(), "helvetica", 12, FONTWEIGHT_BOLD, FONTSLANT_ITALIC, FONTENCODING_DEFAULT))

        @quickViewTitle = FXLabel.new(chat_frame, " -N/A- ")
        @quickViewTitle.setFont(FXFont.new(getApp(), "helvetica", 12, FONTWEIGHT_BOLD, FONTSLANT_ITALIC, FONTENCODING_DEFAULT))
        @quickViewSubTitle = FXLabel.new(chat_frame, "")

        frame = FXHorizontalFrame.new(chat_frame, :opts => LAYOUT_CENTER_X|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH, :padding => 0)
        @mre_button = FXButton.new(frame, "Manual Request", ICON_MANUAL_REQUEST_SMALL, nil, 0, :opts => BUTTON_NORMAL|LAYOUT_RIGHT|LAYOUT_FILL_X)
        @mre_button.connect(SEL_COMMAND) {
          open_manual_request_toolkit(@lastViewed) if @lastViewed
        }
        @mre_button.enabled = false

        @fuzz_button = FXButton.new(frame, "Fuzzer", ICON_FUZZER_SMALL, nil, 0, :opts => BUTTON_NORMAL|LAYOUT_RIGHT|LAYOUT_FILL_X)
        @fuzz_button.connect(SEL_COMMAND) {
          openFuzzer(@lastViewed) if @lastViewed
        }
        @fuzz_button.enabled = false
        @bv_button = FXButton.new(frame, "Browser-View", ICON_BROWSER_SMALL, nil, 0, :opts => BUTTON_NORMAL|LAYOUT_RIGHT|LAYOUT_FILL_X)
        @bv_button.connect(SEL_COMMAND) {
          begin
            if @lastViewed and @browserView then
              openBrowser(@lastViewed.request, @lastViewed.response)
            end
          rescue => bang
            puts bang

          end
        }
        @bv_button.enabled = false

        FXHorizontalSeparator.new(chat_frame, :opts => SEPARATOR_GROOVE|LAYOUT_FILL_X)
        #  FXLabel.new(view_menu, "Source:")

        title_frame = FXHorizontalFrame.new(chat_frame, :opts => LAYOUT_FILL_X)
        FXLabel.new(title_frame, "Request:")

        @request_viewer = Watobo::Gui::RequestViewer.new(chat_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        #  @request_viewer = Watobo::Gui::ChatViewer.new(chat_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        # @request_viewer.highlight_style = 1

        FXHorizontalSeparator.new(chat_frame, :opts => SEPARATOR_GROOVE|LAYOUT_FILL_X)
        title_frame = FXHorizontalFrame.new(chat_frame, :opts => LAYOUT_FILL_X)
        FXLabel.new(title_frame, "Response:")

        #fxViewButton = FXButton.new(title_frame, "View", nil, nil, 0, FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT)
        #fxViewButton.connect(SEL_COMMAND, method(:onViewResponse))

        # @response_viewer = Watobo::Gui::ChatViewer.new(chat_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        @response_viewer = Watobo::Gui::ResponseViewer.new(chat_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y, :padding => 0)
        # @response_viewer.highlight_style = 2

        #===================================================================
        # L O G I N F O
        #===================================================================
        logFrame = FXVerticalFrame.new(@switcher, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN)
        FXLabel.new(logFrame, "Eventlist:", :opts => LAYOUT_FILL_X)
        sunkenlogFrame = FXHorizontalFrame.new(logFrame, :opts => FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y)
        @logText = FXText.new(sunkenlogFrame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|TEXT_AUTOSCROLL)

        # DASHBOARD#
        @dashboard = Dashboard.new(@switcher)

        # FINDING INFORMATION
        frame = FXVerticalFrame.new(@switcher, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_GROOVE)
        @finding_info = FindingInfo.new(frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)

        # PLUGIN-BOARD
        @pluginboard = PluginBoard.new(@switcher)

        if @foption_nopix.checked? then @doctype_TableFilter.concat(@fext_pix);end
        if @foption_nodocs.checked? then @doctype_TableFilter.concat(@fext_docs);end
        if @foption_nojs.checked? then @doctype_TableFilter.concat(@fext_javascript);end
        if @foption_nocss.checked? then @doctype_TableFilter.concat(@fext_style);end
      end

      # !!!
      #  TODO: FXRUBY-Bug???
      #  If splash screen is shown app will crash on close :(
      # !!!
      # def create
      #   super
      # splash = FXSplashWindow.new(interface, ICON_SPLASH, SPLASH_SHAPED|SPLASH_DESTROY|SPLASH_OWNS_ICON, 2000)
      #splash = FXSplashWindow.new(@app, ICON_SPLASH, SPLASH_SIMPLE, 2000)
      # splash.execute
      # splash.create()
      #splash.show(PLACEMENT_SCREEN)
      # show(PLACEMENT_SCREEN)
      #show(PLACEMENT_MAXIMIZED)

      # end
      private

      def subscribeProject(project)

        project.subscribe(:new_chat){ |c|
          Thread.new { addChat(c)}
        }

        project.subscribe(:new_finding){ |f|
          Thread.new {
            addFinding(f)
          }
        }
        
        project.subscribe(:settings_changed){
          saveProjectSettings(project)
          saveSessionSettings(project)
        }
      end

      def subscribeSitesTree()
        @sites_tree.subscribe(:add_site_to_scope){ |site|
          @active_project.addToScope(site)
        }

        @sites_tree.subscribe(:show_conversation){ |chat_list|
          showConversation()
          @chatTable.showConversation(chat_list)
        }

        @sites_tree.subscribe(:show_chat){ |chat|
          showChat(chat)
        }
      end

      def subscribeFindingsTree()
        @findings_tree.subscribe(:add_site_to_scope){ |site|
          @active_project.addToScope(site)
        }

        @findings_tree.subscribe(:delete_domain_filter){ |df|
          @active_project.settings[:domain_filters].delete(df)
          puts "Delete Domain-Filter #{df}"
          updateTreeLists()
        }

        @findings_tree.subscribe(:delete_all_domain_filters) {
          @active_project.settings[:domain_filters].clear
          updateTreeLists()
        }

        @findings_tree.subscribe(:vuln_click){ |v| showVulnerability(v) }

        @findings_tree.subscribe(:finding_click){ |v| showFindingDetails(v) }

        @findings_tree.subscribe(:show_finding_details){ |v| showFindingDetails(v) }

        @findings_tree.subscribe(:open_manual_request){ |v| open_manual_request_toolkit(v) }

      end

      def onClose(sender, sel, event)
        #  puts "! #{Thread.list.length} Threads running"
        response = FXMessageBox.question(self, MBOX_YES_NO, "Finished?", "Are you sure?")
        if response == MBOX_CLICKED_YES
          return 0
        else
          return 1
        end
      end

      def loadProjectSettings(filename=nil)
        settings = nil
        if filename then
          begin
            settings = Hash.new
            settings = Watobo::Utils.loadSettings(filename)
            #            puts settings.to_yaml
          rescue => bang
            puts "!!!ERROR: could not update project settings"
            puts bang
            return false
          end
        end
        return settings
      end

      def loadSessionSettings(filename=nil)
        settings = nil
        if filename then
          begin
            settings = Watobo::Utils.loadSettings(filename)
          rescue => bang
            puts "!!!ERROR: could not load session settings"
            puts bang
            return false
          end
        end
        return settings
      end

      def onExit(sender, sel, item)
        response = FXMessageBox.question(self, MBOX_YES_NO, "Finished?", "Are you sure?")
        if response == MBOX_CLICKED_YES
          getApp().exit(0)
        end
      end

      def onApplyFilterClick(sender,sel,item)
        applyFilter()

      end

      def onClear(sender, sel, item)
        @table_filter.value =""
        @tableFilterFX.handle(self, FXSEL(SEL_UPDATE, 0), nil)
      end

      def onTableClick(sender,sel,item)
        begin
          getApp().beginWaitCursor()
          # purge viewers
          @request_viewer.setText('')
          @response_viewer.setText('')
          row = item.row

          chatid = @chatTable.getRowText(row).to_i
          @chatTable.selectRow(row, false)
          # @logText.appendText("selected ID: (#{chatid})\n")
          @active_project.chats.each do |chat|
            if chat.id == chatid then
              showChat(chat)
            end
          end
        rescue => bang
          puts "!!!ERROR: onTableClick"
          puts bang
          puts "!!!"
        ensure
          getApp().endWaitCursor()
        end
      end

      def onMenuProxy(sender,sel,item)
        if @active_project
          puts @active_project.settings.to_yaml
          proxy_dialog = Watobo::Gui::ProxyDialog.new(self, @active_project.settings[:forwarding_proxy])
          if proxy_dialog.execute != 0 then
            proxy_prefs = proxy_dialog.getProxyPrefs
            @active_project.setProxyOptions(YAML.load(YAML.dump(proxy_prefs)))
            # puts "* current proxy:\n> #{@active_project.getCurrentProxy()}"
            saveDefaultSettings(:forwarding_proxy => proxy_prefs)
            saveProjectSettings(@active_project)
            setStatus(@active_project)
           puts "* current proxy:\n> #{@active_project.getCurrentProxy().to_yaml}"
          end
        else
          FXMessageBox.information(self,MBOX_OK,"No Project Defined", "Create Project First")
        end
      end

      def onCmdShowPreferences_UNUSED(sender, sel, ptr)
        prefDialog = FXpreferences.new(self)
        #  prefDialog.prefs[:source_path].value = @main.prefs[:source_path]
        #  prefDialog.prefs[:source_type].value  = @main.prefs[:source_type]

        @main.prefs.each do |key, value|
          if prefDialog.prefs[key]
            prefDialog.prefs[key].value = @main.prefs[key]
          end
        end

        if prefDialog.execute != 0

          @logText.appendText("Preferences changed...\n")
          prefDialog.prefs.each do |key, value|
            @main.prefs[key]=prefDialog.prefs[key].value

          end

        end
      end

    end # Class End

  end
end