# .
# requesteditor.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
module Watobo
  module Gui
    
    class SimpleTextView < FXVerticalFrame
      
      include Watobo::Gui::Utils
      
      attr :textbox
      attr_accessor :style  
      attr_accessor :max_len
      
      
      public
      
      def rawRequest()
        @textbox.text
      end
      
      def setFont(font_type=nil, size=nil)
        new_size = size.nil? ? GUI_REGULAR_FONT_SIZE : size
        new_font_type = font_type.nil? ? "helvetica" : font_type
        new_font = FXFont.new(getApp(), new_font_type, new_size)
        new_font.create
        @textbox.font = new_font
      end
      
      def initialize(owner, opts)
        
        @logger = (defined? owner.logger) ? owner.logger : nil
        @pattern_matches = []
        @text = ""
        
        super(owner, opts)
        # Construct some hilite styles
        @style = 1 # default style
        @max_len = 0
        
        @small_font = FXFont.new(getApp(), "helvetica", GUI_SMALL_FONT_SIZE)
        @small_font.create
        
        @big_font = FXFont.new(getApp(), "helvetica", GUI_REGULAR_FONT_SIZE)
        @big_font.create
        
        @last_button_pressed = SEL_TYPE_HIGHLIGHT
        
        # Construct some hilite styles
        hs_green = FXHiliteStyle.new
        hs_green.normalForeColor = FXRGBA(255,255,255,255) #FXColor::Red
        hs_green.normalBackColor = FXRGBA(0,255,0,1)   # FXColor::White
        hs_green.style = FXText::STYLE_BOLD
        
        hs_red = FXHiliteStyle.new
        hs_red.normalForeColor = FXRGBA(255,255,255,255) #FXColor::Red
        hs_red.normalBackColor = FXRGBA(255,0,0,1)   # FXColor::White
        hs_red.style = FXText::STYLE_BOLD
        
        
        
        
        # @req_builder = FXText.new(req_editor, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
        @textbox = FXText.new(self, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
        # Enable the style buffer for this text widget
        @textbox.styled = true
        # Set the styles
        @textbox.hiliteStyles = [ hs_green, hs_red]
        
        @textbox.editable = false
        
        @textbox.textStyle |= TEXT_WORDWRAP
        
        @textbox.connect(SEL_RIGHTBUTTONRELEASE) do |sender, sel, event|
          
          unless event.moved?
            FXMenuPane.new(self) do |menu_pane|
              
              pos = @textbox.selStartPos
              len = @textbox.selEndPos - pos
              
              selection = @textbox.extractText(pos, len)
              addStringInfo(menu_pane, sender)
              addDecoder(menu_pane, sender)
              addEncoder(menu_pane, sender) if @textbox.editable?
              
              
              
              FXMenuSeparator.new(menu_pane)
              FXMenuCaption.new(menu_pane,"- Copy -")
              FXMenuSeparator.new(menu_pane)
              copyText = FXMenuCommand.new(menu_pane,"copy text: #{selection}", nil, @textbox, FXText::ID_COPY_SEL)
              
              FXMenuSeparator.new(menu_pane)
              FXMenuCaption.new(menu_pane,"- Transcoder -")
              FXMenuSeparator.new(menu_pane)
              send2transcoder = FXMenuCommand.new(menu_pane,"send to transcoder")
              send2transcoder.connect(SEL_COMMAND) {
                t = FXTranscoder.new(FXApp.instance, selection)
                t.create
                t.show(Fox::PLACEMENT_SCREEN)
              }
              FXMenuSeparator.new(menu_pane)
              target = FXMenuCheck.new(menu_pane, "word wrap" )              
              target.check = ( @textbox.textStyle & TEXT_WORDWRAP > 0 ) ? true : false
              target.connect(SEL_COMMAND) { |sender, sel, item|
                if sender.checked?                    
                  @textbox.textStyle |= TEXT_WORDWRAP 
                else
                  @textbox.textStyle ^= TEXT_WORDWRAP
                end
              }
              
              target = FXMenuCheck.new(menu_pane, "big font" )              
              target.check = ( @textbox.font == @small_font ) ? false : true                
              target.connect(SEL_COMMAND) { |sender, sel, item|                
                if sender.checked?                    
                  @textbox.font = @big_font                    
                else
                  @textbox.font = @small_font                  
                end
                
              }
              
              
              
              menu_pane.create
              menu_pane.popup(nil, event.root_x, event.root_y)
              app.runModalWhileShown(menu_pane)
              
            end            
          end
        end
      end
      
      def editable=(e)
        @textbox.editable = e
      end
      
      def editable?()
        @textbox.editable?
      end      
      
      def setText(text)
        @text = normalizeText(text)
        showText(@text)
      end
      
      
      def filter(pattern)        
        #dummy = @textbox.to_s.split(/\n/)
        dummy = @text.split(/\n/)
        @textbox.setText('')
        filtered = []
        dummy.each do |line|
          begin
            if line =~ /#{pattern}/i then
              filtered.push line
            end
          rescue => bang
            puts 
            puts bang
            pattern = Regexp.quote(pattern)
            retry
          end
        end
        showText(filtered.join("\n"))
      end
      
      def highlight(pattern)
        sindex = nil
        eindex = nil
        
        dummy = @textbox.to_s
        #remove previous highlighting
        @textbox.setText(dummy)
        
        matchPattern(pattern)
        
        # puts "* found pattern #{pattern} #{@pattern_matches.length} times"
        @pattern_matches.each do |start, len|
          begin
            @textbox.changeStyle(start, len, @style)
          rescue => bang
            puts "outch"
            puts bang
          end
        end
        
        return @pattern_matches
      end
      
      def makeMatchVisible(match_index)
        mi = match_index ? match_index : 0
        if @pattern_matches[mi] then
          pos = @pattern_matches[mi][0] 
          len =@pattern_matches[mi][1] 
          @textbox.makePositionVisible(pos + len)
          @textbox.makePositionVisible(pos)
        end
      end
      
      private
      
      # returns an array of matches. each match consists of two values [start, len]
      def matchPattern(pattern)
        sindex = nil
        eindex = nil
        
        pos = 0
        @pattern_matches.clear
        
        loop do
          sindex, eindex = @textbox.findText(pattern, pos, SEARCH_EXACT|SEARCH_IGNORECASE|SEARCH_FORWARD)
          
          sindex, eindex = @textbox.findText(pattern, pos, :flags => SEARCH_REGEX|SEARCH_IGNORECASE|SEARCH_FORWARD) if not sindex
          
          sindex, eindex = @textbox.findText(Regexp.quote(pattern),pos, :flags => SEARCH_REGEX|SEARCH_IGNORECASE|SEARCH_FORWARD) if not sindex
          
          break if not sindex or sindex.length == 0
          
          pos = eindex.last+1
          sindex.length.times do |i|
            start = sindex[i]
            len = eindex[i] - sindex[i]
            @pattern_matches.push [ start, len] if start >= 0  
          end
          
          break if sindex.last < 0
          
        end
        
        return @pattern_matches
      end
      
      def showText(text)
        if @max_len > 0 and @max_len < text.length
          text = text[0..@max_len] + "\n---8<--- WATOBO ---8<---\n* PRESS RESET TO SEE FULL RESPONSE *" 
        end
        atext = []
        text.split(/\n/).each do |line|
          atext.push line.gsub(/[^[:print:]]/,'.')
        end
        @textbox.setText(atext.join("\n"))
      end
      
      def normalizeText_UNUSED(text)
        dummy = [] 
        n = ""
        if text.class.to_s =~ /String/i then 
          text.split("\n").each do |l|
            if @max_len == 0 or dummy.join.length < @max_len then
              n = l.strip
              n.gsub!(/[^[:print:]]/,'.')
              dummy.push n
            else
              break
            end
          end
        elsif text.class.to_s =~ /Array/i then
          begin
            text.headers.each do |h|              
              n = h.strip
              dummy.push n
            end
            dummy.push ""
            text.body.each do |b|
              b.split("\n").each do |l|
                if @max_len == 0 or dummy.join.length < @max_len then
                  n = l.strip
                  n.gsub!(/[^[:print:]]/,'.')
                  dummy.push n
                else
                  break
                end
              end
            end
          rescue => bang
            text.each do |h|              
              n = h.strip
              dummy.push n
            end
            #dummy = text
          end
          
        else
          return nil
        end
        
        return dummy.join("\n")
      end
      
      def normalizeText(text)
        t = text
        t = text.join.gsub(/\r/,"") if text.is_a? Array
        
        dummy = t.split(/\n/)
        dummy.map! do |x|
          x.strip!
          x.gsub!(/[^[:print:]]/,'.')
          x
        end
        return dummy.join("\n")
      end
      
      def log(text,e)
        #   @logger.addError "Not a valid expression! \n#{e}"
        t = Time.now  
        now = t.strftime("%m/%d/%Y @ %H:%M:%S")  
        
        puts "[#{now}] #{text}"
        if e then
          puts e
          puts e.backtrace
        end
      end
      
      
      
    end
    
    
    
    class RequestEditor < SimpleTextView
      
      def initialize(owner, opts)
        super(owner, opts)
        
        @textbox.textStyle -= TEXT_WORDWRAP
        
        @textbox.editable = true
        @markers = []
        
        @textbox.connect(SEL_RIGHTBUTTONRELEASE) do |sender, sel, event|
          unless event.moved?
            FXMenuPane.new(self) do |menu_pane|
              addStringInfo(menu_pane, sender)
              addDecoder(menu_pane, sender)
              addEncoder(menu_pane, sender)
              menu_pane.create
              menu_pane.popup(nil, event.root_x, event.root_y)
              app.runModalWhileShown(menu_pane)
            end
            
          end
        end
        
        @textbox.connect(SEL_REPLACED, method(:onTextChanged))
        @textbox.connect(SEL_DELETED, method(:onTextChanged))
        
        
      end
      
      def parseRequest
        pattern="(%%[^%]*%%)"
        request = @textbox.to_s
        
        begin
          # puts new_request
          expr = ''
          new_request = ''
          request.split(/\n/).each do |line|
            #puts line.unpack("H*")
            new_line = line
            if line =~ /(#{pattern})/ then
              match = $1
              expr = match.gsub(/%%/,"")
              # puts "Executing: #{expr}"
              result = eval(expr)
              new_line = ''
              if result.is_a? File
                data = result.read
                result.close
                
              elsif result.is_a? String
                data = result
              elsif result.is_a? Array
                data = result.join
              else
                log("!!!WATOBO - expression must return String or File !!!",'')
              end
              
              s = line.index(match)
              
              new_line = line[0..s-1] if s > 0
              new_line += data
              new_line += line[s + match.length..-1]
            end
            new_request += new_line + "\n"       
            #puts new_request
          end
        rescue SyntaxError => e
          log("(#{expr}) Not a valid expression!", e)
        rescue LocalJumpError => e
          log("(#{expr}) Not a valid expression!", e)
        rescue => e
          log("(#{expr}) Not a valid expression!", e)
        rescue => bang
          puts bang
          puts bang.backtrace
        end
        
        #   puts new_request
        
        return Watobo::Utils.text2request(new_request) 
        
      end    
      
      
      
      def onTextChanged(sender, sel, changed)
        begin
          dummy = changed.ins
          dummy += changed.del
          pos = changed.pos
          
          pattern = "(%%[^%]*%%)"
          if dummy =~ /%/ then 
            
            @markers = highlight(pattern)
            @textbox.setCursorPos(pos)
          else
            @markers.each do |start, len|
              if pos >= start and pos <= start + len then
                @markers = highlight(pattern)
                @textbox.setCursorPos(pos)
                break
              end
            end  
          end
          
        end
      rescue => bang
        puts "!!!ERROR: onTextChanged"
        puts bang
      end
      
      
      
    end
    
    # -> module Watobo::Gui
  end
end