# .
# possible_login.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
module Watobo
  module Modules
    module Passive
      
      
      class Possible_login < Watobo::Core::PassiveCheck
        
        
        def initialize(project)
          @project = project
          super(project)
          
          @info.update(
                       :check_name => 'Detect Logins',    # name of check which briefly describes functionality, will be used for tree and progress views
          :description => "Detect possible and also unencrypted logins.",   # description of checkfunction
          :author => "Andreas Schmidt", # author of check
          :version => "0.9"   # check version
          )
          
          @finding.update(
                          :threat => 'If login credentials are sent over an unencrypted channel, an attacker may eavesdrop these information.'        # thread of vulnerability, e.g. loss of information
          
          )
          
          @check_name = "Detect Logins"
          @description = "maybe usefull?"
          
          
          @possible_login_patterns=%w[ (username) (password) (passwd) (pass) (uid) (userid) ]
        end
        
        def do_test(chat)
          begin
            #  puts "running module: #{Module.nesting[0].name}"
            all_parms = chat.request.post_parms
            if all_parms
              # puts all_parms
            #  resource = "/" + chat.request.resource
              all_parms.each do |parm|
                @possible_login_patterns.each do |pattern|
                  #  puts "Testing pattern #{pattern} on postparms\r\n#{parm}"
                  if parm =~ /#{pattern}/i
                    match = $1
                    
                    addFinding(
                               :class => "Logins",    # vulnerability class, e.g. Stored XSS, SQL-Injection, ...
                    :type => FINDING_TYPE_HINT,         # FINDING_TYPE_HINT, FINDING_TYPE_INFO, FINDING_TYPE_VULN 
                    :check_pattern => "#{parm}", 
                    :chat => chat,
                    :title => "#{chat.request.path_ext}"
                    #:debug => true
                    )
                    # check for unecrypted transfer
                    
                    if not chat.request.proto =~ /https/i
                      addFinding(
                                 :class => "Unencrypted Logins",    # vulnerability class, e.g. Stored XSS, SQL-Injection, ...
                      :type => FINDING_TYPE_VULN,         # FINDING_TYPE_HINT, FINDING_TYPE_INFO, FINDING_TYPE_VULN 
                      :check_pattern => "#{chat.request.proto}",
                      :chat => chat,
                      :rating => VULN_RATING_HIGH,
                      :title => "#{chat.request.path_ext}"
                     # :debug => true
                      )
                    end
                    return true
                  end
                  
                  
                end
                
              end
            end
          rescue
            puts "ERROR!! #{Module.nesting[0].name}"
          end
        end
        
      end
    end
  end
end