# .
# sslchecker.rb
#   
# Copyright 2010 by siberas, http://www.siberas.de
# 
# This file is part of WATOBO (Web Application Tool Box)
#        http://watobo.sourceforge.com
# 
# WATOBO is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation version 2 of the License.
# 
# WATOBO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with WATOBO; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# .
begin
  require '../lib/templates/plugin.rb'
  # require '../lib/httpparser.rb'
  # require '../lib/qShapers.rb'
rescue LoadError
  puts "could not load #{Dir.getwd}"
end
module Watobo
  module Plugin
    module Sslchecker
      
      class Check < Watobo::Core::ActiveCheck
        attr :cipherlist
        
        def initialize(project)
          super(project)
          
          @result = Hash.new
          
          @info.update(
                       :check_name => 'SSL-Checker',    # name of check which briefly describes functionality, will be used for tree and progress views
          :description => "Test applikation for supportes SSL Ciphers.",   # description of checkfunction
          :author => "Andreas Schmidt", # author of check
          :version => "0.9"   # check version
          )
          
          @finding.update(
                          :threat => 'Attacks on weak encryption ciphers which may lead loss of privacy',        # thread of vulnerability, e.g. loss of information
          :class => "SSL Ciphers",    # vulnerability class, e.g. Stored XSS, SQL-Injection, ...
          :type => FINDING_TYPE_VULN,         # FINDING_TYPE_HINT, FINDING_TYPE_INFO, FINDING_TYPE_VULN
          :rating => VULN_RATING_LOW 
          )
          
          ctx = OpenSSL::SSL::SSLContext.new()
          @cipherlist = Array.new
          ctx.ciphers="eNULL" # because ALL don't include Null-Ciphers!!! 
          ctx.ciphers.each do |c|
            @cipherlist.push c[0]
          end
          
          ctx.ciphers="ALL"
          ctx.ciphers.each do |c|
            @cipherlist.push c[0]
          end
        end
        
        def reset()
          @result.clear
        end
        
        def generateChecks(chat)
          begin
            checker = proc {
              
              test_request = nil
              test_response = nil
              # !!! ATTENTION !!!
              # MAKE COPY BEFORE MODIFIYING REQUEST 
              test = chat.copyRequest
              
              @cipherlist.each do |c|
                ctx = OpenSSL::SSL::SSLContext.new()
                ctx.ciphers = c
                cypher = ctx.ciphers.first
                bits = cypher[2].to_i
                algo = cypher[0]
                test_request, test_response = doRequest(test, :ssl_cipher => c)
                if test_request and test_response
                  notify(:cipher_checked, algo, bits, true)
                  if bits < 128
                   
                    addFinding(  test_request, test_response,
                    :test_item => "#{algo}#{bits}",
                               #:proof_pattern => "#{match}", 
                               :chat => chat,
                               :title => "[#{algo}] - #{bits} Bit"
                    )
                  end
                else
                  notify(:cipher_checked, algo, bits, false)
                  #              puts "!!! ERROR: #{c}"
                end
              end
              return test_request, test_response
            }
            yield checker
          rescue => bang
            puts "!error in module #{Module.nesting[0].name}"
            puts bang
          end
        end
      end
      
      
      class Sslchecker < Watobo::Template::Plugin
        
        def createChat(site)
          chat = nil
          url = "https://#{site}/"
          request = []
          request << "GET " + url +" HTTP/1.1\r\n" 
          request << "Host: " + site + "\r\n" 
          request << "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/x-shockwave-flash, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, */*\r\n" 
          request << "Accept-Language: de\r\n"
          request << "Proxy-Connection: close\r\n" 
          request << "User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)\r\n" 
          request << "\r\n"
          
          chat = Watobo::Core::Chat.new(request, [], :id => 0)
          
          return chat
        end
        
        def onSiteSelect(sender, sel, item)
          @site = item
          #puts item.class
          # @dir_combo.clearItems()
          
          #  @dir_combo.appendItem("nothing selected", nil)
          #if @project then            
          #  @project.listDirs(@site) do |dir|
          # @dir_combo.appendItem(dir.slice(0..35), dir)
          #  end
          #@dir_combo.setCurrentItem(0, true) if @dir_combo.numItems > 0 
          #end
        end
        
        def onDirSelect(sender, sel, item)
          
          @dir = item 
          
        end
        
        def updateView()
          #@project = project
          @site = nil
          @sites_combo.clearItems()
          #@dir_combo.clearItems()
          unless @project.nil? then            
            @project.listSites(:ssl => true).each do |site|
              #puts "Site: #{site}"
              @sites_combo.appendItem(site.slice(0..35), site)
            end
            if @sites_combo.numItems > 0
              @sites_combo.setCurrentItem(0) 
              @site = @sites_combo.getItemData(0)
            else
              logger("No SSL Sites available - you need to visit a SSL Site first!")
            end
          end
          
        end
        
        def logger(message)
          t = Time.now  
          now = t.strftime("%m/%d/%Y @ %H:%M:%S")  
          @log_viewer.insertText(0, "#{now}: #{message}\n")          
        end
        
        def start(sender, sel, item)
          unless @site.nil?
          @result_viewer.setText('')
          
          #puts "Site: #{site}"
          #puts "Directory: #{@dir}"
          chat = createChat(@site)
          checklist = []
          checklist.push @check
          chatlist = []
          chatlist.push chat
          scanner = Watobo::Core::Scanner2.new( chatlist, checklist)
          @pbar.total = @check.cipherlist.length
          @pbar.progress = 0
          @pbar.barColor = 'red' 
          Thread.new {            
            scanner.run(:default => true)
          }
          end
        end
        
        def initialize(owner, project)
          super(owner, "SSL-Plugin", project, :opts => DECOR_ALL,:width=>800, :height=>600)
          @plugin_name = "SSL-Checker"
          @project = project
          @site = nil
          @dir = nil
          
          hs_green = FXHiliteStyle.new
          hs_green.normalForeColor = FXRGBA(255,255,255,255) #FXColor::Red
          hs_green.normalBackColor = FXRGBA(0,255,0,1)   # FXColor::White
          hs_green.style = FXText::STYLE_BOLD
          
          hs_red = FXHiliteStyle.new
          hs_red.normalForeColor = FXRGBA(255,255,255,255) # FXColor::Red
          hs_red.normalBackColor = FXRGBA(255,0,0,1)   # FXColor::White
          hs_red.style = FXText::STYLE_BOLD
          
          
          path = Dir.getwd
          
          mr_splitter = FXSplitter.new(self, LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_VERTICAL|SPLITTER_REVERSED|SPLITTER_TRACKING)
          # top = FXHorizontalFrame.new(mr_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_SIDE_BOTTOM)
          top_frame = FXVerticalFrame.new(mr_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y||LAYOUT_FIX_HEIGHT|LAYOUT_BOTTOM,:height => 500) 
          top_splitter = FXSplitter.new(top_frame, LAYOUT_FILL_X|SPLITTER_HORIZONTAL|LAYOUT_FILL_Y|SPLITTER_TRACKING)
          log_frame = FXVerticalFrame.new(mr_splitter, :opts => LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM,:height => 100)
          
          
          @settings_frame = FXVerticalFrame.new(top_splitter, :opts => LAYOUT_FILL_Y)
          result_frame = FXVerticalFrame.new(top_splitter, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
          log_text_frame = FXVerticalFrame.new(result_frame, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK, :padding=>0)
          @result_viewer = FXText.new(log_text_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
          @result_viewer.styled = true
          # Set the styles
          @result_viewer.hiliteStyles = [ hs_green, hs_red]
          
          FXLabel.new(@settings_frame, "Available Sites:")
          @sites_combo = FXComboBox.new(@settings_frame, 5, nil, 0,
                                        COMBOBOX_STATIC|FRAME_SUNKEN|FRAME_THICK|LAYOUT_SIDE_TOP|LAYOUT_FILL_X)
          #@filterCombo.width =200
          
          @sites_combo.numVisible = 20
          @sites_combo.numColumns = 35
          @sites_combo.editable = false
          @sites_combo.connect(SEL_COMMAND, method(:onSiteSelect))
          begin            
            
            #   FXLabel.new(@settings_frame, "Available Sites:")
            #   @dir_combo = FXComboBox.new(@settings_frame, 5, nil, 0,
            #COMBOBOX_STATIC|FRAME_SUNKEN|FRAME_THICK|LAYOUT_SIDE_TOP|LAYOUT_FILL_X)
            #@filterCombo.width =200
            
            #   @dir_combo.numVisible = 20
            #   @dir_combo.numColumns = 35
            #   @dir_combo.editable = false
            #   @dir_combo.connect(SEL_COMMAND, method(:onDirSelect))
            @pbar = FXProgressBar.new(@settings_frame, nil, 0, LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK|PROGRESSBAR_HORIZONTAL)
            #@pbar.create
            @pbar.connect(SEL_CHANGED) {
              print ":"
            }
            @pbar.progress = 0
            @pbar.total = 0
            @pbar.barColor=0
            @pbar.barColor = 'grey' #FXRGB(255,0,0)
            
            button = FXButton.new(@settings_frame, "start")
            button.connect(SEL_COMMAND, method(:start))
            
            @check = Check.new(@project)
            @check.subscribe(:cipher_checked) { |cipher, bits, result|
              if result == false
                @result_viewer.appendStyledText("checked: #{cipher} - #{bits} - #{result}\n",0)
              elsif bits < 128
                @result_viewer.appendStyledText("checked: #{cipher} - #{bits} - #{result}\n",2)
              else
                @result_viewer.appendStyledText("checked: #{cipher} - #{bits} - #{result}\n",1)
              end
              @pbar.increment(1)
              @pbar.barColor = 'grey' if @pbar.progress == @pbar.total-1
              #puts "#{@pbar.progress} of #{@pbar.total}"
              #     logger
              
            }
            
              @check.subscribe(:new_finding) { |f|
            Thread.new { @project.addFinding(f) }
            }
            
            log_frame_header = FXHorizontalFrame.new(log_frame, :opts => LAYOUT_FILL_X)
            FXLabel.new(log_frame_header, "Logs:" )
            
            
            #log_text_frame = FXHorizontalFrame.new(bottom_frame, :opts => LAYOUT_FILL_X|FRAME_SUNKEN|LAYOUT_BOTTOM)
            log_text_frame = FXVerticalFrame.new(log_frame, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK, :padding=>0)
            @log_viewer = FXText.new(log_text_frame, :opts => LAYOUT_FILL_X|LAYOUT_FILL_Y)
            
            updateView()
          rescue => bang
            puts bang
            puts bang.backtrace
          end
          
          
        end
      end
    end
  end
end


if __FILE__ == $0
  
end